"""
Commands related to torrent files
"""

import functools
import os

from .... import constants, jobs, trackers, utils
from .base import CommandBase


class torrent_add(CommandBase):
    """Add torrent file to BitTorrent client"""

    names = ('torrent-add', 'ta')

    argument_definitions = {
        'CLIENT': {
            'type': utils.argtypes.client,
            'help': ('Case-insensitive BitTorrent client name\n'
                     'Supported clients: ' + ', '.join(utils.btclient.client_names())),
        },
        'TORRENT': {
            'nargs': '+',
            'help': 'Path to torrent file',
        },
        ('--download-path', '-p'): {
            'help': "Parent directory of the torrent's content",
            'metavar': 'DIRECTORY',
        },
    }

    @functools.cached_property
    def jobs(self):
        return (
            jobs.torrent.AddTorrentJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                torrent_files=self.args.TORRENT,
                btclient=utils.btclient.BtClient(
                    name=self.args.CLIENT,
                    url=self.client_options['url'],
                    username=self.client_options['username'],
                    password=self.client_options['password'],
                    download_path=self.download_path,
                    check_after_add=self.client_options['check_after_add'],
                    category=self.client_options.get('category', ''),
                ),
            ),
        )

    @functools.cached_property
    def client_options(self):
        return self.get_options('clients', self.args.CLIENT)

    @functools.cached_property
    def download_path(self):
        if self.args.download_path:
            # Make provided --download-path absolute before doing path
            # translation to get the path the client should use, which may run
            # on a different computer.
            return self.client_options['translate_path'].translate(
                os.path.abspath(self.args.download_path)
            )
        else:
            # Use client's default path
            return None


class torrent_create(CommandBase):
    """
    Create torrent file and optionally add or copy it

    The piece hashes are cached and re-used if possible. This means the torrent
    is not generated for every tracker again. Note that the torrent is always
    generated if files are excluded on one tracker but not on the other. The
    files in each torrent must be identical.
    """

    names = ('torrent-create', 'tc')

    argument_definitions = {}

    subcommand_name = 'TRACKER'
    subcommands = {
        tracker.name: {
            'description': (
                f'Create a torrent file for {tracker.label} '
                'in the current working directory.'
            ),
            'cli': {
                # Default arguments for all tackers
                'CONTENT': {
                    'type': utils.argtypes.content,
                    'help': 'Path to release content',
                },
                ('--exclude-files', '--ef'): {
                    'nargs': '+',
                    'action': 'extend',
                    'metavar': 'PATTERN',
                    'help': ('Glob pattern to exclude from torrent '
                             '(matched case-insensitively against path in torrent)'),
                    'default': [],
                },
                ('--exclude-files-regex', '--efr'): {
                    'nargs': '+',
                    'action': 'extend',
                    'metavar': 'PATTERN',
                    'help': ('Regular expression to exclude from torrent '
                             '(matched case-sensitively against path in torrent)'),
                    'type': utils.argtypes.regex,
                    'default': [],
                },
                ('--reuse-torrent', '-t'): {
                    'nargs': '+',
                    'action': 'extend',
                    'metavar': 'TORRENT',
                    'help': ('Use hashed pieces from TORRENT instead of generating '
                             'them again or getting them from '
                             f'{utils.fs.tildify_path(constants.GENERIC_TORRENTS_DIRPATH)}\n'
                             'TORRENT may also be a directory, which is searched recursively '
                             'for a matching *.torrent file.\n'
                             "NOTE: This option is ignored if TORRENT doesn't match CONTENT properly."),
                    'type': utils.argtypes.existing_path,
                    'default': [],
                },
                ('--add-to', '-a'): {
                    'type': utils.argtypes.client,
                    'metavar': 'CLIENT',
                    'help': ('Case-insensitive BitTorrent client name\n'
                             'Supported clients: ' + ', '.join(utils.btclient.client_names())),
                },
                ('--copy-to', '-c'): {
                    'metavar': 'PATH',
                    'help': 'Copy the created torrent to PATH (file or directory)',
                },
                # Custom arguments defined by tracker for this command
                **tracker.TrackerConfig.argument_definitions.get('torrent-create', {}),
            }
        }
        for tracker in trackers.trackers()
    }

    @functools.cached_property
    def tracker_name(self):
        return self.args.subcommand.lower()

    @functools.cached_property
    def tracker(self):
        return trackers.tracker(
            name=self.tracker_name,
            options={
                **self.config['trackers'][self.tracker_name],
                **vars(self.args),
            },
        )

    @property
    def home_directory(self):
        """Create torrent file in current working directory"""
        return '.'

    @functools.cached_property
    def create_torrent_job(self):
        return jobs.torrent.CreateTorrentJob(
            home_directory=self.home_directory,
            cache_directory=self.cache_directory,
            ignore_cache=self.args.ignore_cache,
            content_path=self.args.CONTENT,
            reuse_torrent_path=(
                tuple(self.args.reuse_torrent)
                + tuple(self.config['config']['torrent-create']['reuse_torrent_paths'])
            ),
            tracker=self.tracker,
            exclude_files=(
                tuple(self.args.exclude_files)
                + tuple(self.args.exclude_files_regex)
            ),
            callbacks={
                'announce_url': self.logout_after_getting_announce_url,
            },
        )

    def logout_after_getting_announce_url(self, announce_url):
        # The "announce_url" signal is emitted twice: First we attempt to get
        # the announce URL and the signal callback is called with `Ellipsis`.
        # Then the signal is emitted again with the announce URL.
        if announce_url is not Ellipsis:
            async def logout():
                await self.tracker.logout()

            self.create_torrent_job.add_task(logout())

    @functools.cached_property
    def add_torrent_job(self):
        if self.args.add_to:
            return jobs.torrent.AddTorrentJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                btclient=utils.btclient.BtClient(
                    name=self.args.add_to,
                    url=self.client_options['url'],
                    username=self.client_options['username'],
                    password=self.client_options['password'],
                    download_path=self.download_path,
                    check_after_add=self.client_options['check_after_add'],
                    category=self.client_options.get('category', ''),
                ),
            )

    @functools.cached_property
    def client_options(self):
        return self.get_options('clients', self.args.add_to)

    @functools.cached_property
    def download_path(self):
        # By default, we assume we are running on the same computer as the
        # client, in which case the download_path is identical to the parent
        # directory of the provided CONTENT path.
        #
        # If the client is running on a different computer, we apply a
        # user-configured search-and-replace path translation.
        #
        # In either case, we first make the download_path absolute for easier
        # trouble-shooting and to make the path translation predictable.
        return self.client_options['translate_path'].translate(
            os.path.abspath(utils.fs.dirname(self.args.CONTENT))
        )

    @functools.cached_property
    def copy_torrent_job(self):
        if self.args.copy_to:
            return jobs.torrent.CopyTorrentJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                destination=self.args.copy_to,
            )

    @functools.cached_property
    def jobs(self):
        return (
            self.create_torrent_job,
            self.add_torrent_job,
            self.copy_torrent_job,
        )



import abc

from ... import errors

import logging  # isort:skip
_log = logging.getLogger(__name__)


class TrackerRuleBase(abc.ABC):
    @property
    def reference(self):
        """Short reference to the relevant rule on the tracker's website"""

    required_jobs = ()
    """..."""

    def __init__(self, tracker_jobs):
        self._tracker_jobs = tracker_jobs

    @property
    def tracker_jobs(self):
        """:class:`~.TrackerJobsBase` instance"""
        return self._tracker_jobs

    @property
    def release_name(self):
        """:class:`~.ReleaseName` instance"""
        return self._tracker_jobs.release_name

    @property
    def tracker(self):
        """:class:`~.TrackerBase` instance"""
        return self._tracker_jobs.tracker

    def check_if_possible(self, job_):
        _log.debug('RULE: %s finished', job_)
        if self._required_jobs_finished():
            self.check()

    def _required_jobs_finished(self):
        ok = True
        for job_name in self.required_jobs:
            job = getattr(self.tracker_jobs, job_name)
            if not job.is_finished:
                ok = False
            _log.debug('RULE CHECK: %r: %r -> %r', job.name, job.is_finished, ok)
        return ok

    @abc.abstractmethod
    def check(self):
        """
        ...
        """

class BannedGroup(TrackerRuleBase):

    required_jobs = (
        'release_name_job',
    )

    banned_groups = {}
    """:class:`set` of banned group names"""

    def is_group(self, group_name):
        """
        Return whether `group_name` is equal to the :attr:`~.ReleaseName.group` of
        :attr:`~.TrackerRuleBase.release_name`
        """
        return self.release_name.group.lower() == group_name.lower()

    def check(self):
        self.custom_check()

        # Case-insensitively match group name against `banned_groups`.
        for banned_group in self.banned_groups:
            _log.debug(f'if self.is_group({banned_group!r}): {self.is_group(banned_group)}')
            if self.is_group(banned_group):
                raise errors.BannedGroup(banned_group)

    def custom_check(self):
        """
        Match group name plus some other attributes (e.g. no encodes from a certain group)

        :raise RuleBroken: if the group is banned under specific conditions
        """



"""
Managing callbacks
"""

import asyncio
import collections
import contextlib
import enum

import logging  # isort:skip
_log = logging.getLogger(__name__)


class Signal:
    """Simple callback registry"""

    class Stopped(enum.Flag):
        """Three-way boolean enum: `false`, `true` and `immediately`"""

        false = 0
        true = 1
        immediately = 2

    def __init__(self, id=None, *, signals):
        self.id = id
        self._signals = {}
        self._suspended_signals = set()
        self._emissions = []
        self._emissions_recorded = []
        self._record_signals = []
        self._queues = collections.defaultdict(list)  # List of asyncio.Queue
        self._is_stopped = self.Stopped.false
        for signal in signals:
            self.add(signal)

    @property
    def id(self):
        """Instance ID used for debugging"""
        return self._id

    @id.setter
    def id(self, id):
        if id in ('', None):
            self._id = 'anonymous'
        else:
            self._id = id

    def add(self, signal, *, record=False):
        """
        Create new `signal`

        :param hashable signal: Any hashable object
        :param bool record: Whether emissions of this signal are recorded in
            :attr:`emissions_recorded`

        :raise TypeError: if `signal` is not hashable
        :raise RuntimeError: if `signal` has been added previously
        """
        if not isinstance(signal, collections.abc.Hashable):
            raise TypeError(f'{self._id}: Unhashable signal: {signal!r}')
        elif signal in self._signals:
            raise RuntimeError(f'{self._id}: Signal already added: {signal!r}')
        else:
            self._signals[signal] = []
            if record:
                self.record(signal)

    @property
    def signals(self):
        """Mutable dictionary of signals mapped to lists of callbacks"""
        return self._signals

    def record(self, signal):
        """Record emissions of `signal` in :attr:`emissions`"""
        if signal not in self._signals:
            raise ValueError(f'{self._id}: Unknown signal: {signal!r}')
        else:
            self._record_signals.append(signal)

    @property
    def recording(self):
        """class:`list` of signals that are recorded"""
        return self._record_signals

    def register(self, signal, callback):
        """
        Call `callback` when `signal` is emited

        :param hashable signal: Previously added signal
        :param callable callback: Any callback. The signature depends on the caller of
            :meth:`emit`.

        :raise ValueError: if `signal` was not added first
        :raise TypeError: if `callback` is not callable
        """
        if signal not in self._signals:
            raise ValueError(f'{self._id}: Unknown signal: {signal!r}')
        elif not callable(callback):
            raise TypeError(f'{self._id}: Not a callable: {callback!r}')
        else:
            self._signals[signal].append(callback)

    def emit(self, signal, *args, **kwargs):
        """
        Call callbacks that are registered to `signal` and make any ongoing :meth:`receive_all`
        calls iterate and any ongoing :meth:`receive_one` calls return

        :param hashable signal: Previously added signal

        Any other arguments are passed on to the callbacks.
        """
        if self.is_stopped:
            raise RuntimeError(f'{self._id}: Cannot emit signal {signal!r} after stop() was called')

        if signal not in self._suspended_signals:
            for callback in self._signals[signal]:
                callback(*args, **kwargs)

            # Store emission.
            self._emissions.append((signal, {'args': args, 'kwargs': kwargs}))

            # Store recorded emission separately.
            if signal in self._record_signals:
                self._emissions_recorded.append((signal, {'args': args, 'kwargs': kwargs}))

            # Tell receive_(all|one)() calls about the emission.
            for queue in self._queues[signal]:
                queue.put_nowait((args, kwargs))

    async def receive_all(self, signal, *, only_posargs=False):
        """
        Iterate over ``(args, kwargs)`` from :meth:`emit` calls for `signal`

        This will always produce all emitted signals, both from the past and the future, until
        :meth:`stop` is called. The only exception is when :meth:`stop` is called with
        ``immediately=True``, in which case iteration stops immediately.

        :param hashable signal: Previously added signal
        :param bool only_posargs: Whether to ignore any keyword arguments from the :meth:`emit` call
            and only iterate over positional arguments
        """
        # Create queue that will get signals via emit(). We do this first because we don't want to
        # miss any new emissions while we are yielding old emissions.
        queue = asyncio.Queue()
        self._queues[signal].append(queue)

        # Yield old emissions that were emitted before we were called, but only if emissions should
        # not stop IMMEDIATELY.
        if self.is_stopped is not self.Stopped.immediately:
            for signal_, emission in self.emissions:
                if signal_ == signal:
                    if only_posargs:
                        yield emission['args']
                    else:
                        yield emission['args'], emission['kwargs']

        if not self._is_stopped:
            # Yield new emissions as they appear until stop() is called.
            while True:
                args, kwargs = await queue.get()
                if args is kwargs is None:
                    break
                elif only_posargs:
                    yield args
                else:
                    yield args, kwargs

    async def receive_one(self, signal, *, only_posargs=False):
        """Same as :meth:`receive_all`, but return just one emission"""
        async for emission in self.receive_all(signal, only_posargs=only_posargs):
            return emission
        # stop() was called before any `signal` was emitted
        return () if only_posargs else ((), {})

    def stop(self, *, immediately=False):
        """
        Do not :meth:`emit` any more signals

        Calling this method disallows further calls to :meth:`emit`, iterating over any ongoing
        :meth:`receive_all` calls stops, and any ongoing :meth:`receive_one` calls return `None`

        :param bool immediately: If truthy :meth:`receive_all` calls stop iterating NOW without
            producing any more emissions and :meth:`receive_one` calls return `None`

            Otherwise, any emissions made before :meth`stop` is called are processed normally before
            iteration stops.
        """
        if immediately:
            self._is_stopped = self.Stopped.immediately

            # Dump all items from all queues so receive_all() stops iteration immediately.
            for queues in self._queues.values():
                for queue in queues:
                    try:
                        queue.get_nowait()
                    except asyncio.queues.QueueEmpty:
                        pass
        else:
            self._is_stopped = self.Stopped.true

        # Put termination sentinels into queues.
        args = kwargs = None
        for queues in self._queues.values():
            for queue in queues:
                queue.put_nowait((args, kwargs))

    @property
    def is_stopped(self):
        """
        Whether :meth:`stop` was called (see :attr:`Stopped`)

        This means :meth:`emit` can no longer be called, any ongoing :meth:`receive_all` calls will
        stop iterating, and any ongoing :meth:`receive_one` calls return `None`.
        """
        return self._is_stopped

    @property
    def emissions(self):
        """
        Sequence of :meth:`emit` calls made so far

        Each call is stored as a tuple like this::

            (<signal>, {"args": <positional arguments>,
                        "kwargs": <keyword arguments>})
        """
        return tuple(self._emissions)

    @property
    def emissions_recorded(self):
        """
        Sequence of only recorded :meth:`emit` calls made so far

        Each call is stored as a tuple like this::

            (<signal>, {"args": <positional arguments>,
                        "kwargs": <keyword arguments>})
        """
        return tuple(self._emissions_recorded)

    def replay(self, emissions):
        """:meth:`emit` previously recorded :attr:`emissions`"""
        for signal, payload in emissions:
            self.emit(signal, *payload['args'], **payload['kwargs'])

    @contextlib.contextmanager
    def suspend(self, *signals):
        """
        Context manager that blocks certain signals in its body

        :param signals: Which signals to block

        :raise ValueError: if any signal in `signals` is not registered
        """
        for signal in signals:
            if signal not in self._signals:
                raise ValueError(f'{self._id}: Unknown signal: {signal!r}')

        self._suspended_signals.update(signals)

        try:
            yield
        finally:
            self._suspended_signals.difference_update(signals)



import functools

from .... import utils
from .. import widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PosterJobWidget(JobWidgetBase):

    is_interactive = True

    def setup(self):
        self._activity_indicator = widgets.ActivityIndicator(
            style='class:info',
            extend_width=True,
        )
        self.job.signal.register('running', lambda _: self._activity_indicator.enable())
        self.job.signal.register('finished', lambda _: self._activity_indicator.disable())

        self.job.signal.register('downloading', self.handle_downloading)
        self.job.signal.register('resizing', self.handle_resizing)
        self.job.signal.register('uploading', self.handle_uploading)

    def handle_downloading(self, url):
        self._activity_indicator.format = f'{{indicator}} Downloading {url}'

    def handle_resizing(self, filepath):
        self._activity_indicator.format = f'{{indicator}} Resizing {utils.fs.basename(filepath)}'

    def handle_uploading(self, imghost):
        self._activity_indicator.format = f'{{indicator}} Uploading to {imghost.name}'

    @functools.cached_property
    def runtime_widget(self):
        return self._activity_indicator



"""
Concrete :class:`~.base.TrackerJobsBase` subclass for UHD
"""

import functools
import re
import urllib.parse
from datetime import datetime

import unidecode

from ... import errors, jobs, utils
from ..base import TrackerJobsBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class UhdTrackerJobs(TrackerJobsBase):

    @functools.cached_property
    def jobs_before_upload(self):
        # Don't start poster_job until we have an IMDb ID.
        self.poster_job.prejobs += (self.imdb_job,)

        return (
            # Interactive jobs
            self.imdb_job,
            self.type_job,
            self.year_job,
            self.quality_job,
            self.version_job,
            self.source_job,
            self.codec_job,
            self.hdr_format_job,
            self.tags_job,
            self.poster_job,
            self.trailer_job,
            self.season_job,
            self.automerge_group_job,
            self.scene_check_job,

            # Background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,
            self.description_job,
        )

    @property
    def isolated_jobs(self):
        """
        Sequence of job attribute names (e.g. "description_job") that were
        singled out by the user, e.g. with --only-description
        """
        if self.options.get('only_description', False):
            return self.get_job_and_dependencies(self.description_job)
        else:
            # Activate all jobs in jobs_before/after_upload
            return ()

    @functools.cached_property
    def type_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('type'),
            label='Type',
            precondition=self.make_precondition('type_job'),
            prejobs=(
                self.imdb_job,
            ),
            autodetect=self.autodetect_type,
            autofinish=True,
            options=(
                ('Movie', '0'),
                # ("Music", '1'),  # Not supported
                ('TV', '2'),
            ),
            **self.common_job_args(ignore_cache=True),
        )

    async def autodetect_type(self, _):
        assert self.imdb_job.is_finished, self.imdb_job
        if self.imdb_id:
            type = await self.imdb.type(self.imdb_id)
            if type is utils.release.ReleaseType.movie:
                return 'Movie'
            elif type in (
                    utils.release.ReleaseType.season,
                    utils.release.ReleaseType.episode,
            ):
                return 'TV'

    @property
    def user_confirmed_type(self):
        # Get type from type_job, which got its type from imdb_job, which is
        # guaranteed to be correct because we asked the user.
        if self.type_job.is_finished and self.type_job.output:
            text = self.type_job.output[0]
            return {
                'Movie': utils.release.ReleaseType.movie,
                'TV': utils.release.ReleaseType.season,
            }[text]

    @functools.cached_property
    def year_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('year'),
            label='Year',
            precondition=self.make_precondition('year_job'),
            prejobs=(
                self.imdb_job,
            ),
            text=self.autodetect_year,
            nonfatal_exceptions=(
                errors.RequestError,
            ),
            normalizer=self.normalize_year,
            validator=self.validate_year,
            finish_on_success=True,
            **self.common_job_args(),
        )

    async def autodetect_year(self):
        assert self.imdb_job.is_finished

        json = await self._tracker.get_uhd_info(self.imdb_id)
        year = json.get('year', None)
        _log.debug('Autodetected UHD year: %r', year)
        if year:
            return year

        year = await self.imdb.year(self.imdb_id)
        if year:
            _log.debug('Autodetected IMDb year: %r', year)
            return year

    def normalize_year(self, text):
        return text.strip()

    def validate_year(self, text):
        if not text:
            raise ValueError('Year must not be empty.')
        try:
            year = int(text)
        except ValueError as e:
            raise ValueError('Year must be a number.') from e
        else:
            if not 1800 < year < datetime.now().year + 10:
                raise ValueError('Year is not reasonable.')

    @functools.cached_property
    def season_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('season'),
            label='Season',
            precondition=self.make_precondition('season_job'),
            prejobs=(
                self.imdb_job,
                self.type_job,
            ),
            hidden=self.season_job_is_hidden,
            text=self.autodetect_season,
            normalizer=self.normalize_season,
            validator=self.validate_season,
            finish_on_success=True,
            **self.common_job_args(),
        )

    def season_job_is_hidden(self):
        return self.user_confirmed_type not in (
            utils.release.ReleaseType.season,
            utils.release.ReleaseType.episode,
        )

    async def autodetect_season(self):
        if self.user_confirmed_type in (
                utils.release.ReleaseType.season,
                utils.release.ReleaseType.episode,
        ):
            _log.debug('Autodetected season: %r', self.release_name.only_season)
            if self.release_name.only_season:
                return self.release_name.only_season
            else:
                return None
        else:
            # Empty string for movies. This finishes the job successfully.
            return ''

    def normalize_season(self, text):
        return text.strip()

    def validate_season(self, text):
        if text:
            try:
                season = int(text)
            except ValueError as e:
                raise ValueError('Season must be a number.') from e
            else:
                # NOTE: Season 0 is ok for pilots, specials, etc.
                if not 0 <= season <= 100:
                    raise ValueError('Season is not reasonable.')

    @property
    def season_number(self):
        output = self.get_job_output(self.season_job, slice=0, default=())
        if output:
            return int(output)

    @functools.cached_property
    def quality_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('quality'),
            label='Quality',
            precondition=self.make_precondition('quality_job'),
            autodetect=self.autodetect_quality,
            autofinish=True,
            options=(
                ('mHD', 'mHD'),
                ('720p', '720p'),
                ('1080p', '1080p'),
                ('1080i', '1080i'),
                ('2160p', '2160p'),
                ('Other', 'Others'),
            ),
            focused='Other',
            **self.common_job_args(),
        )

    async def autodetect_quality(self, _):
        resolution = utils.mediainfo.video.get_resolution(self.content_path)
        _log.debug('Autodetecting quality: %s', resolution)
        if resolution in ('2160p', '1080i', '1080p', '720p'):
            return resolution

    @functools.cached_property
    def version_job(self):
        return jobs.custom.CustomJob(
            name=self.get_job_name('version'),
            label='Version',
            precondition=self.make_precondition('version_job'),
            worker=self.autodetect_version,
            # Non-special releases produce no output, which is not an error.
            no_output_is_ok=True,
            **self.common_job_args(ignore_cache=True),
        )

    async def autodetect_version(self, _):
        versions = set()
        for version, is_version in self.version_map.items():
            if not is_version:
                _log.debug('Unsupported autodetection for %r', version)
            elif is_version(self.release_name):
                _log.debug('Autodetected version: %r', version)
                versions.add(version)
        return versions

    version_map = {
        "Director's Cut": lambda release: "Director's Cut" in release.edition,
        'Theatrical': lambda release: 'Theatrical Cut' in release.edition,
        'Extended': lambda release: 'Extended Cut' in release.edition,
        'IMAX': lambda release: 'IMAX' in release.edition,
        'Uncut': lambda release: 'Uncut' in release.edition,
        'TV Cut': None,  # Unsupported
        'Unrated': lambda release: 'Unrated' in release.edition,
        'Remastered': lambda release: 'Remastered' in release.edition,
        '4K Remaster': lambda release: '4k Remastered' in release.edition,
        '4K Restoration': lambda release: '4k Restored' in release.edition,
        'B&W Version': None,  # Unsupported
        'Criterion': lambda release: 'Criterion Collection' in release.edition,
        '2in1': lambda release: '2in1' in release.edition,
        '3in1': lambda release: '3in1' in release.edition,
        'Hybrid': lambda release: 'Hybrid' in release.source,
        '10-bit': lambda release: utils.mediainfo.video.get_bit_depth(release.path, default=None) == 10,
        'Extras': None,  # Unsupported
    }

    # On the website, this is called "Media".
    @functools.cached_property
    def source_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('source'),
            label='Source',
            precondition=self.make_precondition('source_job'),
            autodetect=self.autodetect_source,
            autofinish=True,
            options=(
                ('Blu-ray', 'Blu-ray'),
                ('Remux', 'Remux'),
                ('Encode', 'Encode'),
                ('WEB-DL', 'WEB-DL'),
                ('WEBRip', 'WEBRip'),
                ('HDRip', 'HDRip'),
                ('HDTV', 'HDTV'),
                ('Other', 'Others'),
            ),
            focused='Other',
            **self.common_job_args(ignore_cache=True),
        )

    async def autodetect_source(self, _):
        for source, is_source in self.source_map.items():
            if is_source(self.release_name):
                _log.debug('Autodetected source: %r', source)
                return source

    source_map = {
        'Remux': lambda release: 'Remux' in release.source,
        'Encode': lambda release: any(
            source in release.source
            for source in (
                'BluRay',
                'HD-DVD',
            )
        ),
        'WEB-DL': lambda release: 'WEB-DL' in release.source,
        'WEBRip': lambda release: 'WEBRip' in release.source,
        'HDTV': lambda release: 'HDTV' in release.source,
        # Not sure what "HDRip" is exactly and how to detect it.
        'HDRip': lambda release: 'Rip' in release.source,
    }

    @functools.cached_property
    def codec_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('codec'),
            label='Codec',
            precondition=self.make_precondition('codec_job'),
            autodetect=self.autodetect_codec,
            autofinish=True,
            options=(
                ('x264', 'x264'),
                ('x265', 'x265'),
                ('x266', 'x266'),
                ('H.264', 'H.264'),  # AVC aka H.264
                ('H.265', 'HEVC'),   # HEVC aka H.265
                ('AV1', 'AV1'),
                ('VC-1', 'VC-1'),
                ('MPEG-2', 'MPEG-2'),
            ),
            **self.common_job_args(ignore_cache=True),
        )

    async def autodetect_codec(self, _):
        for codec, is_codec in self.codec_map.items():
            if is_codec(self.release_name):
                _log.debug('Autodetected video codec: %r', codec)
                return codec

    codec_map = {
        # TODO: Add support for commented-out codecs in utils.video and ReleaseName.
        'x264': lambda release: release.video_format == 'x264',
        'x265': lambda release: release.video_format == 'x265',
        # 'x266': lambda release: release.video_format == 'x266',
        'H.264': lambda release: release.video_format == 'H.264',
        'H.265': lambda release: release.video_format == 'H.265',
        # 'AV1': lambda release: release.video_format == 'AV1',
        # 'VC-1': lambda release: release.video_format == 'VC-1',
        # 'MPEG-2': lambda release: release.video_format == 'MPEG-2',
    }

    @functools.cached_property
    def hdr_format_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('hdr-format'),
            label='HDR',
            precondition=self.make_precondition('hdr_format_job'),
            autodetect=self.autodetect_hdr_format,
            autofinish=True,
            options=(
                ('No', 'No'),
                ('HDR10', 'HDR10'),
                ('HDR10+', 'HDR10+'),
                ('Dolby Vision', 'DoVi'),
            ),
            **self.common_job_args(),
        )

    async def autodetect_hdr_format(self, _):
        for hdr_format, is_hdr_format in self.hdr_format_map.items():
            if is_hdr_format(self.release_name):
                _log.debug('Autodetected HDR format from %r: %r', self.release_name.hdr_format, hdr_format)
                return hdr_format

    hdr_format_map = {
        'Dolby Vision': lambda release: 'DV' in release.hdr_format,
        'HDR10+': lambda release: 'HDR10+' in release.hdr_format,
        'HDR10': lambda release: 'HDR10' in release.hdr_format,
        'No': lambda release: release.hdr_format == '',
    }

    @functools.cached_property
    def tags_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('tags'),
            label='Tags',
            precondition=self.make_precondition('tags_job'),
            prejobs=(
                self.imdb_job,
            ),
            text=self.autodetect_tags,
            nonfatal_exceptions=(
                errors.RequestError,
            ),
            finish_on_success=True,
            normalizer=self.normalize_tags,
            validator=self.validate_tags,
            **self.common_job_args(),
        )

    async def autodetect_tags(self):
        assert self.imdb_job.is_finished

        json = await self._tracker.get_uhd_info(self.imdb_id)
        # Tags are comma-separated and may contain entities like "&eacute;". We
        # split them to a list. Keep in mind that the tags may be an empty
        # string, and `"".split(",")` returns `[""]`.
        return utils.html.as_text(json.get('tag', ''))

    max_tags_length = 200

    def normalize_tags(self, text):
        # Remove whitespace around tags and empty tags,
        seq = [
            tag.strip()
            # Deduplicate while maintaining order.
            for tag in dict.fromkeys(text.split(','))
            if tag.strip()
        ]

        # Ensure maximum length of `max_tag_length`.
        while seq and len(','.join(seq)) > self.max_tags_length:
            del seq[-1]
        return '\n'.join(seq)

    min_tags_count = 3

    def validate_tags(self, text):
        seq = [tag for tag in text.split('\n') if tag]
        if len(seq) < self.min_tags_count:
            raise ValueError(f'At least {self.min_tags_count} tags are required.')

    async def get_poster_from_tracker(self):
        await self.imdb_job.wait_finished()
        json = await self._tracker.get_uhd_info(self.imdb_id)
        poster = json.get('photo', None)
        _log.debug('Poster from UHD: %r', poster)
        if poster:
            return {
                'poster': poster,
                'width': None,
                'height': None,
                'imghosts': (),
                'write_to': None,
            }

    @functools.cached_property
    def trailer_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('trailer'),
            label='Trailer',
            precondition=self.make_precondition('trailer_job'),
            prejobs=(
                self.imdb_job,
            ),
            text=self.autodetect_trailer,
            nonfatal_exceptions=(
                errors.RequestError,
            ),
            normalizer=self.normalize_trailer,
            validator=self.validate_trailer,
            finish_on_success=True,
            **self.common_job_args(ignore_cache=bool('trailer' in self.options)),
        )

    async def autodetect_trailer(self):
        # Get trailer from user.
        if self.options.get('trailer', ''):
            return self.options['trailer']

        # Get trailer from UHD.
        assert self.imdb_job.is_finished
        json = await self._tracker.get_uhd_info(self.imdb_id)
        trailer_id = json.get('trailer', None)
        if trailer_id:
            return trailer_id

        # Default to no trailer.
        return ''

    def normalize_trailer(self, text):
        try:
            return 'https://youtu.be/' + self.get_youtube_id(text)
        except ValueError:
            # Keep invalid URL so the user can fix it.
            return text

    def validate_trailer(self, text):
        if text:
            # Raise ValueError if no YouTube ID is found.
            self.get_youtube_id(text)

    _YOUTUBE_ID_NO_DEFAULT = object()

    def get_youtube_id(self, url, default=_YOUTUBE_ID_NO_DEFAULT):
        """
        Find YouTube video ID in `url`

        :param url: Any YouTube URL
        :param default: Return value if no YouTube ID can be found in `url`

            If not provided, raise :class:`ValueError` if no YouTube ID can be
            found in `url`.
        """
        url_ = urllib.parse.urlparse(url)
        if not url_.hostname:
            if re.search(r'^([a-zA-Z0-9_-]+)$', url):
                # `url` is YouTube ID.
                return url

        elif (
                re.search(r'youtu(?:be|)\.[a-z]{2,3}', url_.hostname)
                and (
                    (match := re.search(r'\bv=([a-zA-Z0-9_-]+)\b', url_.query))
                    or (match := re.search(r'^/(?:embed/|v/|watch/|)([a-zA-Z0-9_-]+)$', url_.path))
                )
        ):
            # Found YouTube ID in `url`.
            return match.group(1)

        if default is self._YOUTUBE_ID_NO_DEFAULT:
            # No default value provided - raise exception.
            raise ValueError('Not a YouTube ID or URL.')
        else:
            return default

    @functools.cached_property
    def description_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            precondition=self.make_precondition('description_job'),
            prejobs=(
                self.screenshots_job,
                self.upload_screenshots_job,
                self.mediainfo_job,
            ),
            text=self.generate_description,
            hidden=True,
            finish_on_success=True,
            read_only=True,
            **self.common_job_args(ignore_cache=True),
        )

    image_host_config = {
        'common': {'thumb_width': 350},
    }

    def generate_description(self):
        screenshots = self.generate_description_screenshots()
        parts = (
            self.generate_description_nfo(),
            (
                f'[center]{screenshots}[/center]'
                if screenshots else
                ''
            ),
            self.promotion_bbcode,
        )
        return '\n\n'.join(part for part in parts if part)

    def generate_description_screenshots(self):
        assert self.upload_screenshots_job.is_finished
        return self.make_screenshots_grid(
            screenshots=self.upload_screenshots_job.uploaded_images,
            columns=2,
            horizontal_spacer='   ',
            vertical_spacer='\n\n',
        )

    def generate_description_nfo(self):
        nfo = self.read_nfo(strip=True)
        if nfo:
            return (
                '[spoiler=NFO]'
                + '[code]'
                + nfo
                + '[/code]'
                + '[/spoiler]'
            )

    @functools.cached_property
    def automerge_group_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('automerge-group'),
            label='Automerge Group',
            precondition=self.make_precondition('automerge_group_job'),
            prejobs=(
                self.imdb_job,
                self.type_job,
                self.season_job,
            ),
            autodetect=self.autodetect_automerge_group,
            autofinish=True,
            options=(
                ('Yes', True),
                ('No', False),
            ),
            **self.common_job_args(ignore_cache=True),
        )

    async def autodetect_automerge_group(self, _):
        assert self.season_job.is_finished
        if self.season_number is not None:
            # All TV seasons share the same IMDB ID, but they are grouped by season
            # number. From the Uploading Guide v2 in the wiki:
            # > Auto-Merge should always be selected unless you are uploading a
            # > TV season that doesn't have a torrent group yet.
            # https://uhdbits.org/wiki.php?action=article&id=33
            #
            # This implementation searches for the IMDb ID and tries to find a
            # group named ".* Season \d+" in the returned HTML.
            season_group_names = await self._get_season_group_names()
            target_group_name_suffix = f'Season {self.season_number:02d}'.lower()
            for group_name in season_group_names:
                # Example season group names:
                #  - The Title
                #  - The Title Season 03
                #  - The Title Season 03 / The AKA Title
                if re.search(rf'\s+{target_group_name_suffix}\b', group_name):
                    _log.debug('Found existing season group: %r', group_name)
                    return 'Yes'
                else:
                    _log.debug('Existing season group does not match %r: %r',
                               target_group_name_suffix, group_name)

            # Looks like season group doesn't exist yet.
            return 'No'
        else:
            # Movies are always autogrouped by IMDb ID
            return 'Yes'

    async def _get_season_group_names(self):
        await self._tracker.login()
        response = await self._tracker._request(
            method='GET',
            url=self._tracker._torrents_url,
            params={
                'searchstr': self.imdb_id,
            },
            error_prefix='Automerge group check failed',
        )

        group_names = []
        doc = utils.html.parse(response)
        for tag in doc.find_all('a', {'class': 'torrent_name'}):
            group_name = utils.html.as_text(tag).strip()
            group_names.append(unidecode.unidecode(group_name.lower()))
        _log.debug('Existing season groups: %r', group_names)
        return group_names

    release_name_translation = {
        'group': {
            re.compile(r'^NOGROUP$'): 'Unknown',
        },
    }

    @property
    def post_data(self):
        return {
            'submit': 'true',

            # Type ("0" for Movie, "2" for TV)
            'type': self.get_job_attribute(self.type_job, 'choice'),

            # IMDb ID ("tt...")
            'imdbid': self.get_job_output(self.imdb_job, slice=0),

            # Original title
            'title': self.release_name.title,

            # English title
            'OtherTitle': self.release_name.title_aka,

            # Uploading Guide v2 says: "Disregard this field"
            'smalldesc': '',

            # Year
            'year': self.get_job_output(self.year_job, slice=0),

            # Season number or `None` (i.e. no submission of "season") for movie
            'season': self.season_number,

            # Quality (e.g. "1080p")
            'format': self.get_job_attribute(self.quality_job, 'choice'),

            # Group
            'team': self.release_name.group,

            # Version (e.g. "Director's Cut")
            'Version': ' / '.join(self.get_job_output(self.version_job)),

            # Source ("Media" on the website) (e.g. "BluRay")
            'media': self.get_job_attribute(self.source_job, 'choice'),

            # Codec (e.g. "x264")
            'codec': self.get_job_attribute(self.codec_job, 'choice'),

            # HDR format (e.g. "HDR10" or "DoVi")
            'hdr': self.get_job_attribute(self.hdr_format_job, 'choice'),

            # "genre_tags" is a side-effect of the dropdown list that lets the
            # user pick tags. This is probably ignored by the server, but we
            # keep it to reduce the bug surface.
            'genre_tags': '---',

            # Tags (We list them one per line for readability, but the server
            # wants them separated by commas.)
            'tags': ','.join(self.get_job_output(self.tags_job, slice=0).split('\n')),

            # Poster URL
            'image': self.get_job_output(self.poster_job, slice=0),

            # Trailer (YouTube ID)
            'trailer': self.get_youtube_id(
                self.get_job_output(self.trailer_job, slice=0),
                default='',
            ),

            # Mediainfo
            'mediainfo': self.get_job_output(self.mediainfo_job, slice=0),

            # Screenshots and release info
            'release_desc': self.get_job_output(self.description_job, slice=0),

            # Group with anything that has the same IMDB ID
            'auto_merge_group': (
                'on'
                if self.get_job_attribute(self.automerge_group_job, 'choice') else
                None
            ),

            # Internal release
            'internal': 'on' if self.options['internal'] else None,

            # No support for exclusive releases (this is intentional)
            # 'exclude': '0',

            # 3D version
            'd3d': '1' if self.options['3d'] else None,

            # Release contains Vietnamese Audio dub
            'vie': '1' if self.options['vie'] else None,

            # Scene release
            'scene': '1' if self.get_job_attribute(self.scene_check_job, 'is_scene_release') else None,

            # Upload anonymously
            'anonymous': '1' if self.options['anonymous'] else None,
        }

    @property
    def post_files(self):
        return {
            'file_input': {
                'file': self.torrent_filepath,
                'mimetype': 'application/x-bittorrent',
            },
        }



"""
Concrete :class:`~.base.TrackerJobsBase` subclass for PTP
"""

import functools
import itertools
import re
from datetime import datetime

import async_lru

from ... import errors, jobs, uis, utils
from ...utils.release import ReleaseType
from ..base import TrackerJobsBase
from . import metadata

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PtpTrackerJobs(TrackerJobsBase):

    @functools.cached_property
    def jobs_before_upload(self):
        # NOTE: Keep in mind that the order of jobs is important for
        #       isolated_jobs: The final job is the overall result, so if
        #       upload_screenshots_job is listed after description_job,
        #       --only-description is going to print the list of uploaded
        #       screenshot URLs.
        return (
            # Common interactive jobs
            self.playlists_job,
            self.imdb_job,
            self.tmdb_job,
            self.type_job,
            self.ptp_group_id_job,
            self.source_job,
            self.scene_check_job,

            # Interactive jobs that only run if movie does not exists on PTP yet
            self.title_job,
            self.year_job,
            self.edition_job,
            self.plot_job,
            self.tags_job,
            self.poster_job,
            self.artists_job,

            # Background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.bdinfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,
            self.audio_languages_job,
            self.subtitle_languages_job,
            self.trumpable_job,
            self.description_job,
        )

    @property
    def isolated_jobs(self):
        """
        Sequence of job attribute names (e.g. "title_job") that were singled
        out by the user, e.g. with --only-title
        """
        if self.options.get('only_description', False):
            return self.get_job_and_dependencies(self.description_job)
        else:
            # Activate all jobs in jobs_before/after_upload
            return ()

    @functools.cached_property
    def type_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('type'),
            label='Type',
            precondition=self.make_precondition('type_job'),
            prejobs=(
                self.imdb_job,
                self.tmdb_job,
            ),
            autodetect=self.autodetect_type,
            options=metadata.types,
            # If user specified a type, e.g. via CLI, do not make them select it
            # again in the TUI.
            autofinish=bool(self.options.get('type')),
            # Ignore cached type if CLI option --type is provided.
            **self.common_job_args(ignore_cache=bool(self.options.get('type'))),
        )

    def autodetect_type(self, _):
        # User explicitly specified type, e.g. via CLI.
        if self.options.get('type'):
            for type, regex in metadata.types.items():
                if regex.search(self.options['type']):
                    return type

        # Miniseries if user selected a TV ID.
        # NOTE: Do not check for ReleaseType.episode because sometimes movies are tracked as
        #       episodes from a collection on IMDb or episodes are allowed to be uploaded as movies.
        _log.debug('Selected IMDb: %r', self.imdb_job.selected)
        if self.imdb_job.selected.get('type') is ReleaseType.season:
            return 'Miniseries'

        _log.debug('Selected TMDb: %r', self.tmdb_job.selected)
        if self.tmdb_job.selected.get('type') is ReleaseType.season:
            return 'Miniseries'

        # Short film if runtime 45 min or less.
        main_video = utils.fs.find_main_video(self.content_path)
        if utils.mediainfo.get_duration(main_video) <= 45 * 60:
            return 'Short Film'

    @functools.cached_property
    def imdb_job(self):
        imdb_job = super().imdb_job
        imdb_job.no_id_ok = True
        return imdb_job

    @functools.cached_property
    def tmdb_job(self):
        tmdb_job = super().tmdb_job
        tmdb_job.no_id_ok = True
        tmdb_job.prejobs += (self.imdb_job,)
        tmdb_job.precondition = self.make_precondition('tmdb_job', precondition=self.no_imdb_id_available)
        return tmdb_job

    def no_imdb_id_available(self):
        return not bool(self.imdb_id)

    @functools.cached_property
    def audio_languages_job(self):
        return jobs.custom.CustomJob(
            name=self.get_job_name('audio-languages'),
            label='Audio Languages',
            precondition=self.make_precondition('audio_languages_job'),
            worker=self.autodetect_audio_languages,
            no_output_is_ok=True,
            catch=(
                errors.ContentError,
            ),
            **self.common_job_args(ignore_cache=True),
        )

    async def autodetect_audio_languages(self, job):
        audio_languages = utils.mediainfo.audio.get_audio_languages(
            self.content_path,
            default='?',
            exclude_commentary=True,
        )
        _log.debug('Audio languages: %r', audio_languages)
        for language in audio_languages:
            self.audio_languages_job.add_output(language)

    @functools.cached_property
    def subtitle_languages_job(self):
        return jobs.custom.CustomJob(
            name=self.get_job_name('subtitle-languages'),
            label='Subtitle Languages',
            precondition=self.make_precondition('subtitle_languages_job'),
            worker=self.autodetect_subtitle_languages,
            no_output_is_ok=True,
            **self.common_job_args(ignore_cache=True),
        )

    async def autodetect_subtitle_languages(self, job):
        # `TrackerJobsBase.subtitles` from our parent class provides all kinds of subtitles,
        # e.g. from .srt, .idx/.sub files and VIDEO_TS or BDMV directories.
        subtitle_languages = [
            subtitle.language or '?'
            for subtitle in self.subtitles
        ]
        _log.debug('Subtitle languages: %r', subtitle_languages)
        for language in subtitle_languages:
            self.subtitle_languages_job.add_output(language)

        # If there are any subtitle tracks without a language tag, tell the user
        # to manually check the subtitle boxes on the website.
        if '?' in subtitle_languages:
            self.subtitle_languages_job.warn(
                "Some subtitle tracks don't have a language tag.\n"
                'Please add any missing subtitle languages manually\n'
                'on the website after uploading.'
            )

    @functools.cached_property
    def trumpable_job(self):
        return jobs.custom.CustomJob(
            name=self.get_job_name('trumpable'),
            label='Trumpable',
            precondition=self.make_precondition('trumpable_job'),
            prejobs=(
                self.audio_languages_job,
                self.subtitle_languages_job,
            ),
            worker=self.autodetect_trumpable,
            no_output_is_ok=True,
            **self.common_job_args(ignore_cache=True),
        )

    async def autodetect_trumpable(self, job):
        reasons = []

        if (
                self.options.get('hardcoded_subtitles', None) in (True, False)
                # User specified "Hardcoded subtitles".
                and self.options['hardcoded_subtitles']
        ):
            reasons.append(metadata.TrumpableReason.HARDCODED_SUBTITLES)
        # Autodetection of hardcoded subtitles would go here.

        if self.options.get('no_english_subtitles', None) in (True, False):
            # User specified "No English subtitles".
            if self.options['no_english_subtitles']:
                reasons.append(metadata.TrumpableReason.NO_ENGLISH_SUBTITLES)
        else:
            # Autodetect "No English subtitles"
            assert self.audio_languages_job.is_finished
            assert self.subtitle_languages_job.is_finished
            audio_languages = self.audio_languages_job.output
            subtitle_languages = self.subtitle_languages_job.output
            if (
                    # Not a silent film?
                    audio_languages
                    # No properly tagged English audio track?
                    and 'en' not in audio_languages
                    # English subtitles are required.
                    and 'en' not in subtitle_languages
            ):
                # No English subtitles found. If there are any audio or
                # subtitle tracks that could be English, ask the user.
                prompt = self.get_trumpable_no_english_subtitles_prompt(audio_languages, subtitle_languages)
                if prompt:
                    self.trumpable_job.add_prompt(prompt)
                else:
                    # Every audio and subtitle track has a language tag and
                    # none of them are English.
                    reasons.append(metadata.TrumpableReason.NO_ENGLISH_SUBTITLES)

        _log.debug('Trumpable reasons from autodetection: %r', reasons)
        return reasons

    def get_trumpable_no_english_subtitles_prompt(self, audio_languages, subtitle_languages):
        audio_languages = self.audio_languages_job.output
        subtitle_languages = self.subtitle_languages_job.output
        question = ''

        if '?' in audio_languages and '?' in subtitle_languages:
            question = (
                'Does this release have an English audio track or\n'
                'English subtitles for the main audio track?\n'
                '(Commentary and the like do not count.)'
            )

        elif '?' in audio_languages:
            question = (
                'Does this release have an English audio track?\n'
                '(Commentary and the like do not count.)'
            )

        elif '?' in subtitle_languages:
            question = 'Does this release have English subtitles for the main audio track?'

        if question:
            return uis.prompts.RadioListPrompt(
                question=question,
                options=(
                    ('Yes', None),
                    ('No', metadata.TrumpableReason.NO_ENGLISH_SUBTITLES),
                ),
                callbacks=(
                    self.trumpable_no_english_subtitles_prompt_callback,
                ),
            )

    def trumpable_no_english_subtitles_prompt_callback(self, answer):
        _log.debug('Trumpable reasons from user: %r', answer)
        _text, reason = answer
        if reason:
            self.trumpable_job.add_output(reason)

    @functools.cached_property
    def description_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            precondition=self.make_precondition('description_job'),
            prejobs=self.get_job_and_dependencies(
                self.playlists_job,
                self.mediainfo_job,
                self.bdinfo_job,
                self.screenshots_job,
                self.upload_screenshots_job,
            ),
            text=self.generate_description,
            read_only=True,
            hidden=True,
            finish_on_success=True,
            **self.common_job_args(ignore_cache=True),
        )

    async def generate_description(self):
        original_release_name = (
            '[size=4][b]'
            + utils.fs.basename(
                utils.fs.strip_extension(self.content_path)
            )
            + '[/b][/size]'
        )

        # For each video, list mediainfo(s) and bdinfo(s), followed by screenshots.
        bbcode_per_video = []
        for info in self.video_info.values():
            # Mediainfo and BDInfo reports zipped together.
            reports = '\n'.join(
                f'[mediainfo]{report}[/mediainfo]'
                for reports in itertools.zip_longest(
                        info.get('mediainfos', ()),
                        (
                            bdinfo.quick_summary
                            for bdinfo in info.get('bdinfos', ())
                        ),
                        fillvalue=None,
                )
                for report in reports
                if report is not None
            )

            # Screenshots.
            screenshots = '\n'.join(
                f'[img={url}]'
                for url in info.get('screenshot_urls', ())
            )

            # `reports` and/or `screenshots` may be empty.
            merged = ''.join((reports, screenshots)).strip()
            if merged:
                bbcode_per_video.append(merged)

        return (
            original_release_name
            + '\n\n'
            + '\n\n[hr]\n'.join(bbcode_per_video)
        ).strip()

    # Create screenshots/mediainfo/bdinfo from each video file or playlist.
    document_all_videos = True

    @functools.cached_property
    def screenshots_job(self):
        # screenshots_job uses screenshots_count, which uses type_job to determine if we're dealing
        # with a miniseries (`screenshots_from_episode`) or something else
        # (`screenshots_from_movie`), so type_job must finish first before we can start.
        screenshots_job = super().screenshots_job
        screenshots_job.prejobs += (self.type_job,)
        return screenshots_job

    @functools.cached_property
    def screenshots_count(self):
        """
        How many screenshots to make

        Return :attr:`options`\\ ``["screenshots_count"]`` it it exists and is truthy. This value
        should be explicitly set by the user, e.g. via a CLI argument or GUI element.

        Otherwise, return a coroutine function that waits for :attr:`type_job`. If the user selected
        "Miniseries", return :attr:`options`\\ ``["screenshots_from_episode"]``. If not, return
        :attr:`options`\\ ``["screenshots_from_movie"]``.
        """
        # CLI option, GUI widget, etc
        if self.options.get('screenshots_count'):
            return self.options['screenshots_count']

        async def get_screenshots_count():
            await self.type_job.wait_started()
            await self.type_job.wait_finished()
            if self.type_job.output == ('Miniseries',):
                return self.options['screenshots_from_episode']
            else:
                return self.options['screenshots_from_movie']

        return get_screenshots_count

    @functools.cached_property
    def ptp_group_id_job(self):
        return jobs.custom.CustomJob(
            name=self.get_job_name('ptp-group-id'),
            label='PTP Group ID',
            precondition=self.make_precondition('ptp_group_id_job'),
            prejobs=self.get_job_and_dependencies(
                self.imdb_job,
            ),
            worker=self.get_ptp_group_id,
            catch=(
                errors.RequestError,
            ),
            **self.common_job_args(),
        )

    async def get_ptp_group_id(self, _):
        # Get group ID from PTP by IMDb ID.
        assert self.imdb_job.is_finished, self.imdb_job
        group_id = await self.tracker.get_ptp_group_id_by_imdb_id(self.imdb_id)
        if group_id:
            _log.debug('PTP group ID via IMDb ID: %r -> %r', self.imdb_id, group_id)
            return group_id
        else:
            # If there is no existing PTP group, ask the user. Usually this
            # happens if PTP doesn't know this movie yet. The user can just
            # press enter to create a new PTP group.
            #
            # But it's also possible that PTP has a group for this movie while
            # it has no IMDb ID.
            self.ptp_group_id_job.info = 'Enter PTP group ID or nothing if this movie does not exist on PTP.'
            ptp_group_id = await self.ptp_group_id_job.add_prompt(
                uis.prompts.TextPrompt()
            )
            _log.debug('PTP group ID from user: %r', ptp_group_id)
            return ptp_group_id.strip()

    @property
    def ptp_group_id(self):
        """
        PTP group ID if :attr:`ptp_group_id_job` is finished and group ID
        was found, `None` otherwise
        """
        if self.ptp_group_id_job.is_finished and self.ptp_group_id_job.output:
            return self.ptp_group_id_job.output[0]

    def ptp_group_does_not_exist(self):
        """
        Whether no releases of the user-selected IMDb ID already exist on PTP

        :attr:`ptp_group_id_job` must be finished when this is called.

        This is used as a :attr:`~.JobBase.precondition` for jobs that are only
        needed if the server doesn't have any releases for this IMDb ID yet.
        """
        assert self.ptp_group_id_job.is_finished, self.ptp_group_id_job
        return not self.ptp_group_id

    @async_lru.alru_cache
    async def get_movie_metadata(self):
        """
        Wrapper around :meth:`~.PtpTrackerJobs.get_movie_metadata` that defaults to from
        :attr:`~.TrackerJobsBase.tmdb`

        If no IMDb ID was selected by the user or if the metadata from PTP is incomplete, try to add
        values from TMDb.
        """
        # Get PTP-normalized metadata by IMDb ID. If the IMDb ID is falsy, we get a metadata dict
        # with empty values.
        metadata = await self.tracker.get_movie_metadata(self.imdb_id)
        _log.debug('PTP metadata:')
        for k, v in metadata.items():
            _log.debug(f'  * {k} = {v!r}')

        if self.tmdb_id:
            # If there is no IMDb ID or the PTP metadata is incomplete or whatever, update missing
            # information with information from TMDb. Some movies are on TMDb but not on IMDb.

            async def update(key, method_name):
                if not metadata[key]:
                    method = getattr(self.tmdb, method_name)
                    try:
                        metadata[key] = await method(self.tmdb_id)
                    except errors.RequestError as e:
                        _log.debug(f'Ignoring {e!r} from {method_name}({self.tmdb_id!r})')

            await update('title', 'title_original')
            await update('plot', 'summary')
            await update('year', 'year')
            await update('poster', 'poster_url')
            # TODO: Somehow translate TMDb genres (TmdbApi.genres()) to PTP tags
            # (ptp.metadata.tags).
            # await update('tags', 'genres')
            await update('countries', 'countries')

            _log.debug(f'Updated metadata with TMDb ID {self.tmdb_id}:')
            for k, v in metadata.items():
                _log.debug(f'  * {k} = {v!r}')

        return metadata

    @functools.cached_property
    def title_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('title'),
            label='Title',
            precondition=self.make_precondition(
                'title_job',
                # Don't run this job if PTP already knows this movie.
                precondition=self.ptp_group_does_not_exist,
            ),
            prejobs=self.get_job_and_dependencies(
                # We need to wait for the PTP group ID to become available
                # before we can check if it exists.
                self.ptp_group_id_job,
                # fetch_title() needs the IMDb ID for get_movie_metadata().
                self.imdb_job,
            ),
            text=self.fetch_title,
            nonfatal_exceptions=(
                # Raised when get_movie_metadata() cannot find IMDb ID.
                errors.RequestError,
            ),
            normalizer=self.normalize_title,
            validator=self.validate_title,
            **self.common_job_args(),
        )

    async def fetch_title(self):
        assert self.imdb_job.is_finished, self.imdb_job
        # Fill in title from release name until it is loaded from PTP.
        self.title_job.text = self.release_name.title
        metadata = await self.get_movie_metadata()
        # Default to `None` to use title from release name.
        return metadata['title'] or None

    def normalize_title(self, text):
        return text.strip()

    def validate_title(self, text):
        if not text:
            raise ValueError('Title must not be empty.')

    @functools.cached_property
    def year_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('year'),
            label='Year',
            precondition=self.make_precondition(
                'year_job',
                # Don't run this job if PTP already knows this movie.
                precondition=self.ptp_group_does_not_exist,
            ),
            prejobs=self.get_job_and_dependencies(
                # We need to wait for the PTP group ID to become available
                # before we can check if it exists.
                self.ptp_group_id_job,
                # fetch_year() needs the IMDb ID for get_movie_metadata().
                self.imdb_job,
            ),
            text=self.fetch_year,
            nonfatal_exceptions=(
                # Raised when get_movie_metadata() cannot find IMDb ID.
                errors.RequestError,
            ),
            normalizer=self.normalize_year,
            validator=self.validate_year,
            **self.common_job_args(),
        )

    async def fetch_year(self):
        assert self.imdb_job.is_finished, self.imdb_job
        # Fill in year from release name until it is loaded from PTP.
        self.year_job.text = self.release_name.year
        metadata = await self.get_movie_metadata()
        # Default to `None` to use year from release name.
        return metadata['year'] or None

    def normalize_year(self, text):
        return text.strip()

    def validate_year(self, text):
        if not text:
            raise ValueError('Year must not be empty.')
        try:
            year = int(text)
        except ValueError as e:
            raise ValueError('Year is not a number.') from e
        else:
            if not 1800 < year < datetime.now().year + 10:
                raise ValueError('Year is not reasonable.')

    @functools.cached_property
    def edition_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('edition'),
            label='Edition',
            precondition=self.make_precondition('edition_job'),
            text=self.autodetect_edition,
            finish_on_success=True,
            **self.common_job_args(ignore_cache=True),
        )

    async def autodetect_edition(self):
        # List of keys in metadata.editions. The corresponding values are
        # returned in the same order as they are specified in metadata.editions.
        edition_keys = []

        for key, is_edition in self.autodetect_edition_map.items():
            if is_edition(self):
                edition_keys.append(key)

        return ' / '.join(
            metadata.editions[key]
            for key in edition_keys
        )

    autodetect_edition_map = {
        'collection.criterion': lambda self: 'Criterion' in self.release_name.edition,
        # 'collection.masters': lambda self: False,
        # 'collection.warner': lambda self: False,

        'edition.dc': lambda self: "Director's Cut" in self.release_name.edition,
        'edition.extended': lambda self: 'Extended Cut' in self.release_name.edition,
        # 'edition.rifftrax': lambda self: False,
        'edition.theatrical': lambda self: 'Theatrical Cut' in self.release_name.edition,
        'edition.uncut': lambda self: 'Uncut' in self.release_name.edition,
        'edition.unrated': lambda self: 'Unrated' in self.release_name.edition,

        'feature.remux': lambda self: 'Remux' in self.release_name.source,
        'feature.2in1': lambda self: '2in1' in self.release_name.edition,
        # 'feature.2disc': lambda self: False,
        # 'feature.3d_anaglyph': lambda self: False,
        # 'feature.3d_full_sbs': lambda self: False,
        # 'feature.3d_half_ou': lambda self: False,
        # 'feature.3D_half_sbs': lambda self: False,
        'feature.4krestoration': lambda self: '4k Restored' in self.release_name.edition,
        'feature.4kremaster': lambda self: '4k Remastered' in self.release_name.edition,
        'feature.10bit': lambda self: (
            utils.mediainfo.video.get_bit_depth(self.content_path, default=None) == 10
            and 'HDR' not in self.release_name.hdr_format
            and 'DV' not in self.release_name.hdr_format
        ),
        # 'feature.extras': lambda self: False,
        # 'feature.2d3d_edition': lambda self: False,
        'feature.dtsx': lambda self: 'DTS:X' in self.release_name.audio_format,
        'feature.dolby_atmos': lambda self: 'Atmos' in self.release_name.audio_format,
        'feature.dolby_vision': lambda self: 'DV' in self.release_name.hdr_format,
        'feature.hdr10': lambda self: re.search(r'HDR10(?!\+)', self.release_name.hdr_format),
        'feature.hdr10+': lambda self: 'HDR10+' in self.release_name.hdr_format,
        'feature.dual_audio': lambda self: self.release_name.has_dual_audio,
        # 'feature.english_dub': lambda self: False,
        'feature.commentary': lambda self: self.release_name.has_commentary,
    }

    @functools.cached_property
    def tags_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('tags'),
            label='Tags',
            precondition=self.make_precondition(
                'tags_job',
                # Don't run this job if PTP already knows this movie.
                precondition=self.ptp_group_does_not_exist,
            ),
            prejobs=self.get_job_and_dependencies(
                # We need to wait for the PTP group ID to become available
                # before we can check if it exists.
                self.ptp_group_id_job,
                # fetch_tags() needs the IMDb ID for get_movie_metadata().
                self.imdb_job,
            ),
            text=self.fetch_tags,
            nonfatal_exceptions=(
                # Raised when get_movie_metadata() cannot find IMDb ID.
                errors.RequestError,
            ),
            normalizer=self.normalize_tags,
            validator=self.validate_tags,
            **self.common_job_args(),
        )

    async def fetch_tags(self):
        assert self.imdb_job.is_finished, self.imdb_job
        metadata = await self.get_movie_metadata()
        return ', '.join(metadata['tags'])

    def normalize_tags(self, text):
        tags = [tag.strip() for tag in text.split(',')]
        deduped = sorted(dict.fromkeys(tags))
        return ', '.join(tag for tag in deduped if tag)

    def validate_tags(self, text):
        if not text.strip():
            raise ValueError('You must provide at least one tag.')

        tags = []
        for tag in text.split(','):
            tag = tag.strip()
            if tag and tag not in metadata.tags:
                raise ValueError(f'Tag is not valid: {tag}')
            else:
                tags.append(tag)

        if len(self.normalize_tags(text)) > 200:
            raise ValueError('You provided too many tags.')

    @functools.cached_property
    def poster_job(self):
        job = super().poster_job
        # Hide poster_job until imdb_job and ptp_group_id_job are done.
        job.prejobs += (
            self.imdb_job,
            self.ptp_group_id_job,
        )
        return job

    def make_poster_job_precondition(self):
        # Do not run poster_job if there is already a PTP group, meaning the
        # server already has a poster.
        return self.make_precondition('poster_job', precondition=self.ptp_group_does_not_exist)

    async def get_poster_from_tracker(self):
        """Get poster URL from PTP or TMDb"""
        await self.imdb_job.wait_finished()
        metadata = await self.get_movie_metadata()
        if metadata['poster']:
            # Return original poster URL without resizing or re-uploading.
            return {
                'poster': metadata['poster'],
                'width': None,
                'height': None,
                'imghosts': (),
                'write_to': None,
            }

    @functools.cached_property
    def plot_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('plot'),
            label='Plot',
            precondition=self.make_precondition(
                'plot_job',
                # Don't run this job if PTP already knows this movie.
                precondition=self.ptp_group_does_not_exist,
            ),
            prejobs=self.get_job_and_dependencies(
                # We need to wait for the PTP group ID to become available
                # before we can check if it exists.
                self.ptp_group_id_job,
                # fetch_plot() needs the IMDb ID for get_movie_metadata().
                self.imdb_job,
            ),
            text=self.fetch_plot,
            nonfatal_exceptions=(
                # Raised when get_movie_metadata() cannot find IMDb ID.
                errors.RequestError,
            ),
            normalizer=self.normalize_plot,
            validator=self.validate_plot,
            finish_on_success=True,
            **self.common_job_args(),
        )

    async def fetch_plot(self):
        assert self.imdb_job.is_finished, self.imdb_job
        metadata = await self.get_movie_metadata()
        return metadata['plot']

    def normalize_plot(self, text):
        return text.strip()

    def validate_plot(self, text):
        if not text:
            raise ValueError('Plot must not be empty.')

    @functools.cached_property
    def artists_job(self):
        return jobs.custom.CustomJob(
            name=self.get_job_name('artists'),
            label='Artists',
            precondition=self.make_precondition(
                'artists_job',
                precondition=self.artists_job_precondition,
            ),
            prejobs=self.get_job_and_dependencies(
                # These jobs must finish before we can check the precondition.
                self.ptp_group_id_job,
                self.imdb_job,
            ),
            worker=self._get_artists,
            **self.common_job_args(),
        )

    def artists_job_precondition(self):
        # Don't ask user to provide artists if we can get them from PTP or IMDb.
        return (
            not self.ptp_group_id  # PTP does not know this movie.
            and not self.imdb_id   # IMDb does not know this movie.
        )

    async def _get_artists(self, job):
        # Semi-automatic artist entry with artists from TMDb.
        await self._get_artists_from_tmdb(job)

        # Allow the user to add more artists manually.
        if job.output:
            # We found artists on TMDb. Focus "Stop adding artists" so user can just press enter.
            await self._get_artists_from_user(job, focused=-1)
        else:
            # We haven't found artists on TMDb. Focus "Add actor".
            await self._get_artists_from_user(job, focused=0)

    async def _get_artists_from_tmdb(self, job):
        await self.tmdb_job.wait_finished()
        if self.tmdb_id:
            artists = {
                metadata.ArtistImportance.DIRECTOR: await self.tmdb.directors(self.tmdb_id),
                metadata.ArtistImportance.ACTOR: await self.tmdb.cast(self.tmdb_id),
            }
            _log.debug('Artists found on TMDb:')
            for k, v in artists.items():
                _log.debug(' * %r: %r', k, v)

            # Get user confirmation for each artist.
            for importance, persons in artists.items():
                for person in persons:
                    _log.debug(f'Confirming TMDb artist: {importance!r}: {person}')
                    # TODO: Add `role` property to webdbs.common.Person class and provide that along
                    #       with the name so it can be autofilled in the prompt.
                    artist = await self._get_one_artist_from_user(job, importance, name=str(person), role=person.role)
                    # The user can prevent the artist from being added if they enter an empty name, in
                    # which case we get a falsy `artist` (None, empty `dict`, whatever).
                    if artist:
                        self._add_formatted_artist(job, artist)
                    else:
                        return

    async def _get_artists_from_user(self, job, focused):
        # Loop over artist prompts until the user stops the loop.
        while True:
            importance = await self._get_artist_importance(focused=focused)
            _log.debug('Adding artist: %r', importance)
            if importance:
                artist = await self._get_one_artist_from_user(job, importance)
                if artist:
                    self._add_formatted_artist(job, artist)
                else:
                    break
            else:
                # No `importance` means the user wants to stop entering artists.
                break

    async def _get_one_artist_from_user(self, job, importance, *, name='', role=''):
        # Keep asking for the name or other ID until we get a valid one.
        current_name = name
        while True:
            try:
                name, ptpurl = await self._get_artist_name_and_ptpurl(
                    question=f'{importance} name:',
                    name=current_name,
                )
            except errors.RequestedNotFoundError as e:
                _log.debug('Unknown artist: %r', e)
                job.warn(e)
                current_name = e.requested
            except errors.RequestError as e:
                _log.debug('Failed to get artist name: %r', e)
                job.warn(e)
            else:
                break
        _log.debug('Artist name and URL: (%r, %r)', name, ptpurl)
        job.clear_warnings()

        if name and ptpurl:
            # Only actors have a role, and it may be empty.
            if importance == metadata.ArtistImportance.ACTOR:
                role = await self._get_artist_role(question=f'{name} role (optional):', role=role)
                _log.debug('Artist role: %r', role)
            else:
                role = ''
            return {
                'importance': importance,
                'name': name,
                'ptpurl': ptpurl,
                'role': role,
            }

    async def _get_artist_name_and_ptpurl(self, question, name=''):
        self.artists_job.info = (
            'Enter artist name, IMDb link/ID, PTP link/ID '
            'or nothing to stop entering artists.'
        )
        id = await self.artists_job.add_prompt(uis.prompts.TextPrompt(
            question=question,
            text=name,
        ))
        id = id.strip()
        self.artists_job.info = ''

        if not id:
            # User wants to stop entering artists.
            return None, None

        # Get canonical name and PTP ID. This works for IMDb ID/URL and PTP URL.
        try:
            artist = await self.tracker.get_artist_metadata(id)
        except errors.RequestedNotFoundError as e:
            # Create new artist after confirmation or raise "Artist not found".
            artist_name = id
            if (
                    await self.artists_job.add_prompt(uis.prompts.RadioListPrompt(
                        question=f'Create new artist with the name "{artist_name}"?',
                        options=('Yes', 'No'),
                        focused='No',
                    )) == 'Yes'
                    and
                    await self.artists_job.add_prompt(uis.prompts.RadioListPrompt(
                        question=f'Are you sure "{artist_name}" does not exist on PTP or IMDb?',
                        options=('Yes', 'No'),
                        focused='No',
                    )) == 'Yes'
            ):
                artist = await self.tracker.create_artist(artist_name)
            else:
                raise e

        _log.debug('Got artist metadata: %r', artist)
        return artist['name'], artist['url']

    async def _get_artist_importance(self, focused=None):
        _label, enum = await self.artists_job.add_prompt(uis.prompts.RadioListPrompt(
            options=[
                (f'Add {str(importance).lower()}', importance)
                for importance in metadata.ArtistImportance
            ] + [
                ('Stop adding artists', None)
            ],
            focused=focused,
        ))
        return enum

    async def _get_artist_role(self, question, role=''):
        role = await self.artists_job.add_prompt(uis.prompts.TextPrompt(
            question=question,
            text=role,
        ))
        return role.strip()

    def _add_formatted_artist(self, job, artist):
        # IMPORTANT: If you change the output format, make sure to also change the regex in
        # _post_data_add_new_movie_artists.
        if artist.get('role'):
            job.add_output(f'{artist["importance"]}: {artist["name"]} | {artist["role"]} | {artist["ptpurl"]}')
        else:
            job.add_output(f'{artist["importance"]}: {artist["name"]} | {artist["ptpurl"]}')

    @functools.cached_property
    def source_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('source'),
            label='Source',
            precondition=self.make_precondition('source_job'),
            text=self.autodetect_source,
            normalizer=self.normalize_source,
            validator=self.validate_source,
            finish_on_success=True,
            # Ignore cached source if CLI option --source is provided.
            **self.common_job_args(ignore_cache=bool(self.options.get('source'))),
        )

    def autodetect_source(self):
        if self.options.get('source'):
            # Get pre-specified source, e.g. from CLI argument.
            source = self.options['source']
        else:
            # Get source from release name.
            source = self.release_name.source

        # Find autodetected source in valid PTP sources
        for known_source, regex in metadata.sources.items():
            if regex.search(source):
                # Finish the job without prompting the user
                return known_source

        # Let the user fix the autodetected source
        self.source_job.text = source

    def normalize_source(self, text):
        text = text.strip()
        for source, regex in metadata.sources.items():
            if regex.search(text):
                return source
        return text

    def validate_source(self, text):
        if not text or text == 'UNKNOWN_SOURCE':
            raise ValueError('You must provide a source.')
        elif text not in metadata.sources:
            raise ValueError(f'Source is not valid: {text}')

    @property
    def post_data(self):
        post_data = self._post_data_common

        _log.debug('PTP group ID: %r', self.ptp_group_id)
        if self.ptp_group_id:
            _log.debug('Adding movie format to existing group')
            post_data.update(self._post_data_add_movie_format)
        else:
            _log.debug('Creating new movie group')
            post_data.update(self._post_data_add_new_movie)

        return post_data

    @property
    def _post_data_common(self):
        return {
            # Feature Film, Miniseries, Short Film, etc
            'type': self.get_job_attribute(self.type_job, 'choice'),

            # Mediainfo and Screenshots
            'release_desc': self.get_job_output(self.description_job, slice=0),

            # Is not main movie (bool)
            'special': '1' if self.options['not_main_movie'] else None,

            # Is personal rip (bool)
            'internalrip': '1' if self.options['personal_rip'] else None,

            # Is scene Release (bool)
            'scene': '1' if self.get_job_attribute(self.scene_check_job, 'is_scene_release') else None,

            # .nfo file content
            'nfo_text': self.read_nfo(),

            # Upload token from staff
            'uploadtoken': self.options.get('upload_token', None),

            **self._post_data_common_source,
            **self._post_data_common_codec,
            **self._post_data_common_container,
            **self._post_data_common_resolution,
            **self._post_data_common_edition,
            **self._post_data_common_subtitles,
            **self._post_data_common_trumpable,
        }

    @property
    def _post_data_common_source(self):
        return {
            'source': 'Other',
            'other_source': self.get_job_output(self.source_job, slice=0),
        }

    @property
    def _post_data_common_codec(self):
        return {
            # Server-side autodetection
            'codec': '* Auto-detect',
            'other_codec': '',

            # # Custom codec
            # 'codec': 'Other',
            # 'other_codec': self.get_job_output(self.video_codec_job, slice=0),
        }

    @property
    def _post_data_common_container(self):
        return {
            # Server-side autodetection
            'container': '* Auto-detect',
            'other_container': ''

            # # Custom container
            # 'container': 'Other',
            # 'other_container': self.get_job_output(self.container_job, slice=0),
        }

    @property
    def _post_data_common_resolution(self):
        return {
            # Server-side autodetection
            'resolution': '* Auto-detect',
            'other_resolution': '',
        }

        # Custom resolution
        # resolution = self.get_job_output(self.resolution_job, slice=0)
        # # WIDTHxHEIGHT
        # if 'x' in resolution:
        #     width, height = resolution.split('x', maxsplit=1)
        #     return {
        #         'resolution': 'Other',  # Custom resolution
        #         'other_resolution_width': width,
        #         'other_resolution_height': height,
        #     }

        # # 1080p, 720p, etc
        # else:
        #     return {
        #         'resolution': resolution,
        #         'other_resolution': releaseInfo.Resolution,
        #     }

    @property
    def _post_data_common_subtitles(self):
        # Translate BCP47 codes into PTP numeric codes from upload.php
        codes = []
        for sub in self.subtitles:
            ptp_code = None

            if ptp_code is None and sub.region and sub.forced:
                # Language, country and "forced" flag, e.g. "pt-BR (forced)"
                ptp_code = metadata.subtitles.get(f'{sub.language}-{sub.region} (forced)')

            if ptp_code is None and sub.region:
                # Language and "forced" flag, e.g. "pt-BR"
                ptp_code = metadata.subtitles.get(f'{sub.language}-{sub.region}')

            if ptp_code is None and sub.forced:
                # Language and "forced" flag, e.g. "pt (forced)"
                ptp_code = metadata.subtitles.get(f'{sub.language} (forced)')

            if ptp_code is None:
                # Language, e.g. "pt"
                ptp_code = metadata.subtitles.get(sub.language)

            if ptp_code:
                codes.append(ptp_code)

        # Add "No subtitles" if there aren't any or if none of the existing subtitles are supported
        # in metadata.subtitles, i.e. we couldn't find any `codes`.
        if not codes:
            codes.append(metadata.subtitles['No Subtitles'])

        _log.debug('PTP subtitle codes: %r', codes)
        return {'subtitles[]': codes}

    @property
    def _post_data_common_edition(self):
        text = self.get_job_output(self.edition_job, slice=0)
        return {
            # Edition Information ("Director's Cut", "Dual Audio", etc.)
            'remaster': 'on' if text else None,
            'remaster_title': text if text else None,
            # 'remaster_year': ...,
            # 'remaster_other_input': ...,
        }

    @property
    def _post_data_common_trumpable(self):
        # NOTE: Only certain classes can submit trumpable tags. Trumpabale tags
        # from petty users are simply ignored and staff will have to add them.
        return {
            'trumpable[]': [
                # Convert pretty strings to API values
                metadata.TrumpableReason.from_string(string).value
                for string in self.trumpable_job.output
            ],
        }

    @property
    def _post_data_add_movie_format(self):
        # Upload another release to existing movie group
        return {
            'groupid': self.ptp_group_id,
        }

    @property
    def _post_data_add_new_movie(self):
        # Upload movie that is not on PTP yet in any format
        post_data = {
            # IMDb ID (must be 7 characters without the leading "tt")
            'imdb': self.tracker.normalize_imdb_id(self.imdb_id),
            # Release year
            'title': self.get_job_output(self.title_job, slice=0),
            # Release year
            'year': self.get_job_output(self.year_job, slice=0),
            # Movie plot or summary
            'album_desc': self.get_job_output(self.plot_job, slice=0),
            # Genre
            'tags': self.get_job_output(self.tags_job, slice=0),
            # Youtube ID
            # 'trailer': ...,
            # Poster URL
            'image': self.get_job_output(self.poster_job, slice=0),
        }
        post_data.update(self._post_data_add_new_movie_artists)
        return post_data

    @property
    def _post_data_add_new_movie_artists(self):
        # artists_job may not be running at all because artists are handled by
        # the server after the upload via the provided IMDb or PTP ID, so we
        # default to an empty output.
        lines = self.get_job_output(self.artists_job, default=())
        if lines:
            artistnames = []
            artistids = []
            importances = []
            roles = []

            # Possible formats:
            # > {importance}: {name} | {role} | {ptpurl}
            # > {importance}: {name} | {ptpurl}
            line_regex = re.compile(
                r'^(?P<importance>\w+?): '
                r'(?P<name>[^\|]+?) '
                r'(?:\| (?P<role>[^\|]+?) |)'
                r'\| (?P<ptpurl>https://.*)$'
            )

            # PTP URL format:
            # > https://hostname/artist.php?id=123456
            id_regex = re.compile(r'^.*?\bid=(\d+).*$')

            for line in lines:
                _log.debug('OUTPUT: %r', line)
                match = line_regex.search(line)
                _log.debug('match: %r', match)
                if match:
                    # Get ArtistImportance from human-readable string.
                    importance = metadata.ArtistImportance.from_string(match.group('importance'))

                    # Get PTP ID from URL.
                    id = id_regex.sub(r'\1', match.group('ptpurl'))
                    assert all(c in '0123456789' for c in id), id

                    artistnames.append(match.group('name'))
                    importances.append(importance.value)
                    artistids.append(id)
                    roles.append(match.group('role') or '')

                else:
                    raise RuntimeError(f'Unexpected line: {line}')

            # Sanity check: Make sure all lists have the same length.
            artists_count = len(lines)
            for lst in (
                    artistnames,
                    artistids,
                    importances,
                    roles,
            ):
                assert len(lst) == artists_count, (lst, artists_count)

            return {
                'artistnames[]': artistnames,
                'artistids[]': artistids,
                'importances[]': importances,
                'roles[]': roles,
            }

        else:
            return {}



import abc
import functools

from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout.containers import ConditionalContainer, DynamicContainer, HSplit, Window, to_container
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.layout import walk

from ... import prompts
from .. import widgets

import logging  # isort:skip
_log = logging.getLogger(__name__)


class JobWidgetBase(abc.ABC):
    """User-interaction and information display for :class:`~.jobs.JobBase` instance"""

    _empty_widget = Window(
        dont_extend_height=True,
        style='class:info',
    )

    def __init__(self, job, app):
        self._job = job
        self._app = app
        self._prompts = []
        self.setup()

        self.job.signal.register('info', lambda _: self.invalidate())
        self.job.signal.register('warning', lambda _: self.invalidate())
        self.job.signal.register('prompt', self.add_prompt)

        main_widget = HSplit(
            children=[
                # Status information or user interaction
                ConditionalContainer(
                    filter=Condition(lambda: not self.job.is_finished),
                    content=self._prompt_or_runtime_widget,
                ),
                # Ephemeral info message that is only shown while job is running
                ConditionalContainer(
                    filter=Condition(lambda: not self.job.is_finished and bool(self.job.info)),
                    content=self.info_widget,
                ),
                # Warnings
                ConditionalContainer(
                    filter=Condition(lambda: bool(self.job.warnings)),
                    content=self.warnings_widget,
                ),
                # Output
                ConditionalContainer(
                    filter=Condition(lambda: self.job.output),
                    content=self.output_widget,
                ),
                # Errors
                ConditionalContainer(
                    filter=Condition(lambda: bool(self.job.errors)),
                    content=self.errors_widget,
                ),
            ],
        )

        label = widgets.HLabel(
            group='jobs',
            text=self.job.label,
            style='class:label',
            content=main_widget,
        )

        self._container = ConditionalContainer(
            filter=Condition(lambda: self.job.errors or not self.job.hidden),
            content=label,
        )

    @property
    def job(self):
        """Underlying :class:`~.JobBase` instance"""
        return self._job

    @abc.abstractmethod
    def setup(self):
        """
        Called on object creation

        Create widgets and register :attr:`job` callbacks.
        """

    @property
    @abc.abstractmethod
    def runtime_widget(self):
        """
        Interactive or status widget that is displayed while this job is running

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object or
            `None`
        """

    @functools.cached_property
    def _runtime_widget_with_local_keybindings(self):
        if self.runtime_widget:
            container = to_container(self.runtime_widget)
            container.key_bindings = self.keybindings_local
            return container

    @functools.cached_property
    def _prompt_or_runtime_widget(self):
        def get_content():
            if self.current_prompt_widget:
                return self.current_prompt_widget
            elif self._runtime_widget_with_local_keybindings:
                return self._runtime_widget_with_local_keybindings
            else:
                return self._empty_widget

        return DynamicContainer(get_content)

    @property
    def output_widget(self):
        """
        Job :attr:`~.JobBase.output`

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object
        """

        def join_output(job=self.job):
            output = job.output
            if all(
                    (
                        # Maximum length for comma-separated output.
                        len(o) <= 6
                        # Don't use ", " as separator if the output itself
                        # contains ",".
                        and ',' not in str(o)
                    )
                    for o in output
            ):
                joined = ', '.join(output)
            else:
                # Newline-separated output.
                joined = '\n'.join(output)

            # FIXME: If output is empty, prompt-toolkit ignores the
            #        "dont_extend_height" argument. Using a space (0x20) as a
            #        placeholder seems to prevent this issue.
            #
            #        WARNING: The bug also manifests if output is ("",) (i.e. a
            #        non-empty list containing one or more empty strings), which
            #        evaluates as `True`, so `if job.output` doesn't work.
            return joined or ' '

        return Window(
            style='class:output',
            content=FormattedTextControl(join_output),
            dont_extend_height=True,
            wrap_lines=True,
        )

    @property
    def info_widget(self):
        """
        :attr:`~.JobBase.info` that is only displayed while this job is running

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object
        """
        return Window(
            style='class:info',
            content=FormattedTextControl(lambda: str(self.job.info)),
            dont_extend_height=True,
            wrap_lines=True,
        )

    @property
    def warnings_widget(self):
        """
        Any :attr:`~.JobBase.warnings`

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object
        """
        return Window(
            style='class:warning',
            content=FormattedTextControl(lambda: '\n'.join(str(e) for e in self.job.warnings)),
            dont_extend_height=True,
            wrap_lines=True,
        )

    @property
    def errors_widget(self):
        """
        Any :attr:`~.JobBase.errors`

        :return: :class:`~.prompt_toolkit.layout.containers.Window` object
        """
        return Window(
            style='class:error',
            content=FormattedTextControl(lambda: '\n'.join(str(e) for e in self.job.errors)),
            dont_extend_height=True,
            wrap_lines=True,
        )

    def add_prompt(self, prompt):
        self._prompts.append(prompt)
        self._clear_cached_prompt_widget()

    @functools.cached_property
    def current_prompt_widget(self):
        if self._prompts:
            current_prompt = self._prompts[0]
            return self._prompt_widget_factory(
                current_prompt,
                callback=self._handle_prompt_accepted,
            )

    def _prompt_widget_factory(self, prompt, callback):
        if isinstance(prompt, prompts.RadioListPrompt):
            return widgets.RadioList(
                question=prompt.parameters['question'],
                options=prompt.parameters['options'],
                focused=prompt.parameters['focused'],
                on_accepted=callback,
            )
        elif isinstance(prompt, prompts.CheckListPrompt):
            return widgets.CheckList(
                question=prompt.parameters['question'],
                options=prompt.parameters['options'],
                focused=prompt.parameters['focused'],
                on_accepted=callback,
            )
        elif isinstance(prompt, prompts.TextPrompt):
            input_field = widgets.InputField(
                text=prompt.parameters['text'],
                on_accepted=lambda buffer: callback(buffer.text),
                style='class:dialog.text',
            )
            # The question or prompt message is optional, but HLabel always puts
            # space between the question and the input field. This looks like
            # the input field is weirdly indented if the question is empty.
            if prompt.parameters['question']:
                return widgets.HLabel(
                    text=prompt.parameters['question'],
                    content=input_field,
                    style='class:dialog.label',
                )
            else:
                return input_field
        else:
            raise ValueError(f'Unsupported prompt: {type(prompt).__name__!r}: {prompt!r}')

    def _clear_cached_prompt_widget(self):
        try:
            del self.current_prompt_widget
        except AttributeError:
            pass
        self.invalidate()

    def _handle_prompt_accepted(self, result):
        current_prompt = self._prompts[0]
        self._prompts.remove(current_prompt)
        self._clear_cached_prompt_widget()
        current_prompt.set_result(result)

    @functools.cached_property
    def is_interactive(self):
        """
        Whether this job may require user interaction at any point

        Subclasses should hardcode this as a class attribute to `True` or
        `False`. The default implementation tries to guess interactivity by
        finding a focusable widget. This is expensive and can be inaccurate
        because widgets come and go.
        """
        for c in walk(to_container(self._prompt_or_runtime_widget), skip_hidden=True):
            if isinstance(c, Window) and c.content.is_focusable():
                return True
        return False

    @functools.cached_property
    def keybindings_global(self):
        """
        Application-wide :class:`prompt_toolkit.key_binding.KeyBindings` instance

        These keybindings are always active as long as the TUI is running,
        regardless of which widget is focused. This also makes it possible to
        bind keys for unfocusable wigets.

        To keep things tidy and intuitive, all keybindings should be sequences
        of multiple keys. The first key should always start with ``c-<x>`` where
        ``c`` stands for ``Control`` and ``<x>`` is a character that is
        associated with the job's :attr:`~.JobBase.label`.

        For example, :class:`~.CreateTorrentJob` keybindings start with ``c-t``
        and :class:`~.SceneCheckJob` keybindings start with ``c-s``.
        """
        return self._app.key_bindings

    @functools.cached_property
    def keybindings_local(self):
        """
        Widget-wide :class:`prompt_toolkit.key_binding.KeyBindings` instance

        These keybindings are only active as long as this widget is focused. If
        there are conflicts, local keybindings take precedence over
        :attr:`global keybindings <keybindings_global>`.
        """
        return KeyBindings()

    def invalidate(self):
        """Schedule redrawing of the TUI"""
        try:
            del self.is_interactive
        except AttributeError:
            pass

        self._app.invalidate()

    def __pt_container__(self):
        return self._container



"""
Find most recent version
"""

import re

import natsort
from packaging.version import parse as parse_version

from .. import __changelog__, __project_name__, __version__
from . import http

import logging  # isort:skip
_log = logging.getLogger(__name__)


_PYPI_URL = f'https://pypi.org/pypi/{__project_name__}/json'
_REPO_URL = (
    'https://codeberg.org/plotski/'
    f'{__project_name__}/raw/branch/master/{__project_name__}/__init__.py'
)
_MAX_CACHE_AGE = 24 * 3600  # 24 hours
_REQUEST_TIMEOUT = 30


def current_is_prerelease():
    """Whether the currently running release is a prerelease"""
    current = parse_version(__version__)
    return current.is_prerelease


async def get_newer_release():
    """
    Return version of newer release than :attr:`upsies.__version__`

    If there is no newer release, return an empty :class:`str`.

    :raise RequestError: if getting the latest version fails
    """
    current = __version__
    release = await _get_latest_release()
    if parse_version(release) > parse_version(current):
        return release


async def get_newer_prerelease():
    current = __version__
    release = await _get_latest_prerelease()
    if parse_version(release) > parse_version(current):
        return release

    # If there is no newer prerelease, a proper release could've been made after the prerelease we
    # are currently running. We must do a normal release check as well to detect that.
    return await get_newer_release()


async def _get_latest_release():
    response = await http.get(
        url=_PYPI_URL,
        timeout=_REQUEST_TIMEOUT,
        cache=True,
        max_cache_age=_MAX_CACHE_AGE,
    )
    all_versions = natsort.natsorted(response.json()['releases'])
    # PyPI should return the releases sorted by release date (latest last)
    if all_versions:
        return _fix_version(all_versions[-1])


async def _get_latest_prerelease():
    response = await http.get(
        url=_REPO_URL,
        timeout=_REQUEST_TIMEOUT,
        cache=True,
        max_cache_age=_MAX_CACHE_AGE,
    )
    match = re.search(r'^__version__\s*=\s*[\'"]([\w\.]+)[\'"]', response, flags=re.MULTILINE)
    if match:
        return match.group(1)


def _fix_version(version):
    # Python's version parser removes leading zeros from our non-standard date version.
    # We re-add it here for prettification.
    match = re.search(r'^(\d{4})\.(\d+)\.(\d+)(.*)$', version)
    if match:
        year, month, day, pre = match.groups()
        version = f'{year}.{month:0>2}.{day:0>2}{pre}'
    return version


async def get_latest_release_changelog():
    """
    Return changes of the most recent proper release as multiline :class:`str`

    If no proper release can be found, return an empty :class:`str`.

    :raise RequestError: if getting the changelog fails
    """
    changelogs = await _get_recent_changelogs()
    # Find most recent changelog with proper version.
    for version, changelog in changelogs:
        if re.search(r'^\d{4}\.\d+\.\d+$', version):
            return changelog
    return ''


async def get_latest_changelog():
    """Same as :func:`get_latest_release_changelog`, but do not ignore prerelease"""
    changelogs = await _get_recent_changelogs()
    # Always return the first entry if it exists.
    if changelogs:
        return changelogs[0][1]
    else:
        return ''


async def _get_recent_changelogs():
    response = await http.get(
        url=__changelog__,
        timeout=_REQUEST_TIMEOUT,
        cache=True,
        max_cache_age=_MAX_CACHE_AGE,
    )

    version_regex = r'^(\d{4}\..{1,2}\..{1,2})'
    changelogs = []
    for i, changelog in enumerate(re.split(r'\n{2,}', response)):
        # We don't care about old changes.
        if i >= 2:
            break

        if match := re.search(version_regex, changelog, flags=re.MULTILINE):
            version = match.group(1)
            changes = re.sub(version_regex + '.*?\n', r'', changelog, flags=re.MULTILINE).rstrip()
            changelogs.append((version, changes))

    return tuple(changelogs)



"""
Fixed values that do not change during runtime
"""

import os

from xdg.BaseDirectory import xdg_cache_home as XDG_CACHE_HOME
from xdg.BaseDirectory import xdg_config_home as XDG_CONFIG_HOME

from . import __project_name__

DEFAULT_CACHE_DIRECTORY = os.path.join(XDG_CACHE_HOME, __project_name__)
"""Default path cache directory if config option ``cache_directory`` is not set"""

GENERIC_TORRENTS_DIRPATH = os.path.join(DEFAULT_CACHE_DIRECTORY, 'generic_torrents')
"""Path to directory that contains cached torrents for re-using piece hashes"""

CONFIG_DIRECTORYPATH = os.path.join(XDG_CONFIG_HOME, __project_name__)
"""Path to configuration directory"""

CONFIG_FILEPATH = os.path.join(CONFIG_DIRECTORYPATH, 'config.ini')
"""Path to general configuration file"""

TRACKERS_FILEPATH = os.path.join(CONFIG_DIRECTORYPATH, 'trackers.ini')
"""Path to trackers configuration file"""

IMGHOSTS_FILEPATH = os.path.join(CONFIG_DIRECTORYPATH, 'imghosts.ini')
"""Path to image hosting services configuration file"""

CLIENTS_FILEPATH = os.path.join(CONFIG_DIRECTORYPATH, 'clients.ini')
"""Path to BitTorrent clients configuration file"""

VIDEO_FILE_EXTENSIONS = (
    'avi',
    'divx',
    'm2ts',
    'mkv',
    'mp4',
    'mpeg',
    'mpg',
    'mts',
    'ts',
    'vob',
    'wmv',
)
"""Sequence of file extensions to recognize video files"""

GUESSIT_OPTIONS = {
    'expected_title': [
        'The Collector',
        'xXx',
        'TC 2000',
    ],
    'advanced_config': {
    },
}
"""`options` argument for :meth:`.GuessItApi.guessit`"""



"""
Scene release search and check
"""

from .. import errors
from ..utils import fs, predbs, types
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class SceneSearchJob(JobBase):
    """
    Search for scene release

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``search_results``
            Emitted after new search results are available. Registered callbacks
            get a sequence of release names (:class:`str`) as a positional
            argument.
    """

    name = 'scene-search'
    label = 'Scene Search'
    hidden = True
    cache_id = None  # Don't cache output

    def initialize(self, *, content_path, predb=None):
        """
        Set internal state

        :param predb: :class:`~.PredbApiBase` or :class:`~.MultiPredbApi`
        :param content_path: Path to video file or directory that contains a
            video file or release name
        """
        self._predb = predb or predbs.MultiPredbApi()
        self._content_path = content_path
        self.signal.add('search_results')

    async def run(self):
        try:
            results = await self._predb.search(
                query=self._content_path,
                only_existing_releases=False,
            )
        except (errors.RequestError, errors.SceneError) as e:
            _log.debug('Caught %r', e)
            self.error(e)
        else:
            if results:
                for result in results:
                    self.add_output(result)
            else:
                self.error('No results')
            self.signal.emit('search_results', results)


class SceneCheckJob(JobBase):
    """
    Verify scene release name and content

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``ask_release_name``
            Emitted if the user must pick a release name from multiple search
            results. Registered callbacks get a sequence of release names as
            positional argument.

        ``ask_is_scene_release``
            Emitted after we made our best guess and the user must approve or
            override it. Registered callbacks get a
            :class:`~.types.SceneCheckResult` enum as a positional argument.

        ``checked``
            Emitted after :attr:`is_scene_release` is set. Registered callbacks
            get a :class:`~.utils.types.SceneCheckResult` enum as a positional
            argument.
    """

    name = 'scene-check'
    label = 'Scene Check'
    hidden = False
    cache_id = None  # Don't cache output

    @property
    def content_stem(self):
        """Final segment of `content_path` without the file extension"""
        return fs.strip_extension(fs.basename(self._content_path))

    @property
    def is_scene_release(self):
        """
        :class:`~.utils.types.SceneCheckResult` enum or `None` before job is
        finished
        """
        return self._is_scene_release

    def initialize(self, *, content_path, predb=None, force=None):
        """
        Set internal state

        :param predb: :class:`~.PredbApiBase` or :class:`~.MultiPredbApi`
        :param content_path: Path to video file or directory that contains a
            video file or release name
        :param bool force: Predetermined check result; `True` (is scene),
            `False` (is not scene) or `None` (autodetect)
        """
        self._predb = predb or predbs.MultiPredbApi()
        self._content_path = content_path
        self._predetermined_result = None if force is None else bool(force)
        self._is_scene_release = None
        self._check_tasks = []
        self.signal.add('ask_release_name')
        self.signal.add('ask_is_scene_release')
        self.signal.add('checked')
        self.signal.record('checked')
        self.signal.register('checked', lambda is_scene: setattr(self, '_is_scene_release', is_scene))

    async def _catch_errors(self, coro):
        try:
            return await coro
        except errors.RequestError as e:
            # Maybe service is down and we have to ask the user
            self.warn(e)
            self.signal.emit('ask_is_scene_release', types.SceneCheckResult.unknown)
        except errors.SceneError as e:
            # Something like SceneRenamedError or SceneAbbreviatedFilenameError
            self.error(e)

    async def run(self):
        if self._predetermined_result is None:
            # We create a task so we can cancel it if the user makes a decision
            # while we are checking, e.g. by pressing a key.
            self._check_tasks.append(
                self.add_task(
                    self._catch_errors(
                        self._verify()
                    )
                )
            )
        elif self._predetermined_result:
            self.set_result(types.SceneCheckResult.true)
        else:
            self.set_result(types.SceneCheckResult.false)

        # Because this job can be interactive, we must wait for a finalize()
        # call that happens when a decision is made (manually or automatically).
        await self.finalization()

    async def _verify(self):
        _log.debug('Verifying release: %r', self._content_path)

        if await self._predb.is_scene_release(self._content_path) is types.SceneCheckResult.false:
            # Try to get a true negative as fast as possible
            self.set_result(types.SceneCheckResult.false)

        elif predbs.is_mixed_season_pack(self._content_path):
            # We don't want to prompt the user if this is a mixed release
            await self._verify_release()

        else:
            results = await self._predb.search(
                query=self._content_path,
                only_existing_releases=False,
            )
            if self.content_stem in results:
                # Basename of content_path without extension is existing release
                await self._verify_release()
            elif len(results) == 1:
                # Don't prompt user to pick search result if there is only one
                await self._verify_release()
            else:
                # If there are multiple search results, ask the user which one
                # `content_path` is supposed to be
                self.signal.emit('ask_release_name', results)

    def user_selected_release_name(self, release_name):
        """
        Must be called by the UI when the user picked a release name from search
        results

        :param release_name: Scene release name as :class:`str` or any falsy
            value for a non-scene release
        """
        _log.debug('User selected release name: %r', release_name)
        if release_name:
            self._check_tasks.append(
                self.add_task(
                    self._catch_errors(
                        self._verify_release(release_name)
                    )
                )
            )
        else:
            self._handle_scene_check_result(types.SceneCheckResult.false)

    async def _verify_release(self, release_name=None):
        is_scene_release, exceptions = await self._predb.verify_release(self._content_path, release_name)
        self._handle_scene_check_result(is_scene_release, exceptions)

    def _handle_scene_check_result(self, is_scene_release, exceptions=()):
        _log.debug('Handling result: %r: %r', is_scene_release, exceptions)

        warnings = [e for e in exceptions if isinstance(e, errors.SceneMissingInfoError)]
        for e in warnings:
            self.warn(e)

        serious_errors = [e for e in exceptions if not isinstance(e, errors.SceneMissingInfoError)]
        for e in serious_errors:
            self.error(e)

        # Serious errors terminate this job.
        if not serious_errors:
            if is_scene_release in (types.SceneCheckResult.true, types.SceneCheckResult.false):
                self.set_result(is_scene_release)
            else:
                self.signal.emit('ask_is_scene_release', is_scene_release)

    def set_result(self, is_scene_release):
        """
        Make the final decision of whether this is a scene release or not and
        finish this job

        Must be called by the UI in the handler of the signal
        ``ask_is_scene_release``.

        :param is_scene_release: :class:`~.types.SceneCheckResult` enum
        """
        _log.debug('Final scene check decision: %r', is_scene_release)
        self.signal.emit('checked', is_scene_release)

        if is_scene_release is types.SceneCheckResult.true:
            self.add_output('Scene release')
        elif is_scene_release is types.SceneCheckResult.false:
            self.add_output('Not a scene release')
        else:
            self.add_output('May be a scene release')

        # Unblock run() and finish the job.
        self.finalize()

    def stop_checking(self):
        """
        Cancel any ongoing checking tasks

        This method should be called by the UI if the user makes a decision
        while we are trying to autodetect.
        """
        for task in self._check_tasks:
            task.cancel()



"""
Concrete :class:`~.base.TrackerConfigBase` subclass for MTV
"""

import base64

from ... import utils
from ..base import TrackerConfigBase, exclude

MtvImageHost = utils.types.ImageHost(
    allowed=(
        ('imgbox', 'ptpimg', 'dummy')
        if utils.is_running_in_development_environment() else
        ('imgbox', 'ptpimg')
    ),
)


class MtvTrackerConfig(TrackerConfigBase):
    defaults = {
        'base_url': base64.b64decode('aHR0cHM6Ly93d3cubW9yZXRoYW50di5tZQ==').decode('ascii'),
        'username': '',
        'password': '',
        'announce_url': utils.configfiles.config_value(
            value='',
            description='Your personal announce URL. Automatically fetched from the website if not set.',
        ),
        'image_host': utils.configfiles.config_value(
            value=utils.types.ListOf(
                item_type=MtvImageHost,
                default=('imgbox',),
                separator=',',
            ),
            description=(
                'List of image hosting service names. The first service is normally used '
                'with the others as backup if uploading to the first fails.\n'
                + 'Supported services: ' + ', '.join(MtvImageHost.options)
            ),
        ),
        'screenshots_count': utils.configfiles.config_value(
            value=utils.types.Integer(4, min=3, max=10),
            description='How many screenshots to make.',
        ),
        'exclude': (
            exclude.checksums,
            exclude.extras,
            exclude.images,
            exclude.nfo,
            exclude.samples,
            exclude.subtitles,
        ),
        'anonymous': utils.configfiles.config_value(
            value=utils.types.Bool('no'),
            description='Whether your username is displayed on your uploads.',
        ),
    }

    argument_definitions = {
        'submit': {
            ('--imdb', '--im'): {
                'help': 'IMDb ID or URL',
                'type': utils.argtypes.webdb_id('imdb'),
            },
            ('--anonymous', '--an'): {
                'help': 'Hide your username for this submission',
                'action': 'store_true',
                # This must be `None` so it doesn't override the "anonymous"
                # value from the config file. See CommandBase.get_options().
                'default': None,
            },
            ('--screenshots-count', '--ssc'): {
                'help': ('How many screenshots to make '
                         f'(min={defaults["screenshots_count"].min}, '
                         f'max={defaults["screenshots_count"].max})'),
                'type': utils.argtypes.number_of_screenshots(
                    min=defaults['screenshots_count'].min,
                    max=defaults['screenshots_count'].max,
                ),
            },
            ('--screenshots', '--ss'): {
                'help': (
                    'Path(s) to existing screenshot file(s)\n'
                    'Directories are searched recursively.\n'
                    'Precreated screenshots are used in addition to automatically generated screenshots.'
                ),
                'nargs': '+',
                'action': 'extend',
                'type': utils.argtypes.files_with_extension('png'),
                'metavar': 'SCREENSHOT',
            },
            ('--only-description', '--od'): {
                'help': 'Only generate description (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
            ('--only-title', '--ot'): {
                'help': 'Only generate title (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
            ('--ignore-dupes', '--id'): {
                'help': 'Force submission even if the tracker reports duplicates',
                'action': 'store_true',
            },
        },
    }



"""
Concrete :class:`~.base.TrackerConfigBase` subclass for BHD
"""

import base64

from ... import utils
from ..base import TrackerConfigBase, exclude

BhdImageHost = utils.types.ImageHost(
    allowed=(
        ('imgbox', 'ptpimg', 'imgbb', 'dummy')
        if utils.is_running_in_development_environment() else
        ('imgbox', 'ptpimg', 'imgbb')
    ),
)


class BhdTrackerConfig(TrackerConfigBase):
    defaults = {
        'upload_url': base64.b64decode('aHR0cHM6Ly9iZXlvbmQtaGQubWUvYXBpL3VwbG9hZA==').decode('ascii'),
        'announce_url': utils.configfiles.config_value(
            value=base64.b64decode('aHR0cHM6Ly9iZXlvbmQtaGQubWUvYW5ub3VuY2U=').decode('ascii'),
            description='The announce URL without the private passkey.',
        ),
        'announce_passkey': utils.configfiles.config_value(
            value='',
            description=(
                'The private part of the announce URL.\n'
                'Get it from the website: My Security -> Passkey'
            ),
        ),
        'apikey': utils.configfiles.config_value(
            value='',
            description=(
                'Your personal private API key.\n'
                'Get it from the website: My Security -> API key'
            ),
        ),
        'anonymous': utils.configfiles.config_value(
            value=utils.types.Bool('no'),
            description='Whether your username is displayed on your uploads.',
        ),
        'draft': utils.configfiles.config_value(
            value=utils.types.Bool('no'),
            description=(
                'Whether your uploads are stashed under Torrents -> Drafts '
                'after the upload instead of going live.'
            ),
        ),
        'image_host': utils.configfiles.config_value(
            value=utils.types.ListOf(
                item_type=BhdImageHost,
                default=('imgbox',),
                separator=',',
            ),
            description=(
                'List of image hosting service names. The first service is normally used '
                'with the others as backup if uploading to the first fails.\n'
                + 'Supported services: ' + ', '.join(BhdImageHost.options)
            ),
        ),
        'screenshots_count': utils.configfiles.config_value(
            value=utils.types.Integer(4, min=3, max=10),
            description='How many screenshots to make.',
        ),
        'exclude': (
            exclude.checksums,
            exclude.extras,
            exclude.images,
            exclude.nfo,
            exclude.samples,
            exclude.subtitles,
        ),
    }

    argument_definitions = {
        'submit': {
            ('--imdb', '--im'): {
                'help': 'IMDb ID or URL',
                'type': utils.argtypes.webdb_id('imdb'),
            },
            ('--tmdb', '--tm'): {
                'help': 'TMDb ID or URL',
                'type': utils.argtypes.webdb_id('tmdb'),
            },
            ('--anonymous', '--an'): {
                'help': 'Hide your username for this submission',
                'action': 'store_true',
                # This must be `None` so it doesn't override the "anonymous"
                # value from the config file. See CommandBase.get_options().
                'default': None,
            },
            ('--custom-edition', '--ce'): {
                'help': 'Non-standard edition, e.g. "Final Cut"',
                'default': '',
            },
            ('--draft', '--dr'): {
                'help': 'Upload as draft',
                'action': 'store_true',
                # The default value must be None so CommandBase.get_options()
                # doesn't always overwrite the value with the config file value.
                'default': None,
            },
            ('--nfo',): {
                'help': 'Path to NFO file (supersedes any *.nfo file found in the release directory)',
            },
            ('--personal-rip', '--pr'): {
                'help': 'Tag as your own encode',
                'action': 'store_true',
            },
            ('--screenshots-count', '--ssc'): {
                'help': ('How many screenshots to make '
                         f'(min={defaults["screenshots_count"].min}, '
                         f'max={defaults["screenshots_count"].max})'),
                'type': utils.argtypes.number_of_screenshots(
                    min=defaults['screenshots_count'].min,
                    max=defaults['screenshots_count'].max,
                ),
            },
            ('--screenshots', '--ss'): {
                'help': (
                    'Path(s) to existing screenshot file(s)\n'
                    'Directories are searched recursively.\n'
                    'Precreated screenshots are used in addition to automatically generated screenshots.'
                ),
                'nargs': '+',
                'action': 'extend',
                'type': utils.argtypes.files_with_extension('png'),
                'metavar': 'SCREENSHOT',
            },
            ('--special', '--sp'): {
                'help': 'Tag as special episode, e.g. Christmas special (ignored for movie uploads)',
                'action': 'store_true',
            },
            ('--only-description', '--od'): {
                'help': 'Only generate description (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
            ('--only-title', '--ot'): {
                'help': 'Only generate title (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
        },
    }



"""
Concrete :class:`~.base.TrackerBase` subclass for PTP
"""

import asyncio
import math
import re
import urllib

from ... import errors, utils
from ..base import TrackerBase
from .config import PtpTrackerConfig
from .jobs import PtpTrackerJobs

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PtpTracker(TrackerBase):
    name = 'ptp'
    label = 'PTP'
    torrent_source_field = 'PTP'

    setup_howto_template = (
        '{howto.introduction}\n'
        '\n'
        '{howto.next_section}. Login Credentials\n'
        '\n'
        '   {howto.current_section}.1 $ upsies set trackers.{tracker.name}.username USERNAME\n'
        '   {howto.current_section}.2 $ upsies set trackers.{tracker.name}.password PASSWORD\n'
        '\n'
        '{howto.next_section}. Announce URL\n'
        '\n'
        '   The announce URL is required because the passkey is used for 2FA.\n'
        '\n'
        '   {howto.current_section}.1 $ upsies set trackers.{tracker.name}.announce_url ANNOUNCE_URL\n'
        '\n'
        '{howto.next_section}. Screenshots\n'
        '\n'
        '   {howto.current_section}.1 Configure ptpimg.me API\n'
        '       $ upsies upload-images ptpimg --help\n'
        '\n'
        '   {howto.current_section}.2 Specify how many screenshots to make. (optional)\n'
        '       $ upsies set trackers.{tracker.name}.screenshots_from_movie NUMBER_OF_MOVIE_SCREENSHOTS\n'
        '       $ upsies set trackers.{tracker.name}.screenshots_from_episode NUMBER_OF_EPISODE_SCREENSHOTS\n'
        '\n'
        '{howto.autoseed}\n'
        '\n'
        '{howto.reuse_torrents}\n'
        '\n'
        '{howto.upload}\n'
    )

    TrackerConfig = PtpTrackerConfig
    TrackerJobs = PtpTrackerJobs

    @property
    def _base_url(self):
        return self.options['base_url']

    @property
    def _ajax_url(self):
        return urllib.parse.urljoin(self._base_url, '/ajax.php')

    @property
    def _artist_url(self):
        return urllib.parse.urljoin(self._base_url, '/artist.php')

    @property
    def _logout_url(self):
        return urllib.parse.urljoin(self._base_url, '/logout.php')

    @property
    def _upload_url(self):
        return urllib.parse.urljoin(self._base_url, '/upload.php')

    @property
    def _torrents_url(self):
        return urllib.parse.urljoin(self._base_url, '/torrents.php')

    @property
    def _announce_url(self):
        announce_url = self.options['announce_url']
        if not announce_url:
            raise errors.AnnounceUrlNotSetError(tracker=self)
        else:
            return self.options['announce_url']

    @property
    def _passkey(self):
        # Needed for logging in with ajax.php
        match = re.search(r'.*/([a-zA-Z0-9]+)/announce', self._announce_url)
        if match:
            return match.group(1)
        else:
            raise RuntimeError(f'Failed to find passkey in announce URL: {self._announce_url}')

    async def _request(self, method, *args, error_prefix='', **kwargs):
        # Because HTTP errors (e.g. 404) are raised, we treat RequestErrrors as
        # normal response so we can get the message from the HTML.
        try:
            # `method` is "GET" or "POST"
            response = await getattr(utils.http, method.lower())(
                *args,
                user_agent=True,
                follow_redirects=False,
                **kwargs,
            )
        except errors.RequestError as e:
            response = e

        # Get error from regular exception (e.g. "Connection refused") or the
        # HTML in response.
        try:
            self._maybe_raise_error(response)
        except errors.RequestError as e:
            # Prepend error_prefix to explain the general nature of the error.
            if error_prefix:
                raise errors.RequestError(f'{error_prefix}: {e}') from e
            else:
                raise e
        else:
            return response

    def _maybe_raise_error(self, response_or_request_error):
        # utils.http.get()/post() raise RequestError on HTTP status codes, but
        # we want to get the error message from the response text.
        # _maybe_raise_error_from_*() handle Response and RequestError.
        self._maybe_raise_error_from_json(response_or_request_error)
        self._maybe_raise_error_from_html(response_or_request_error)

        # If we got a RequestError and we didn't find an error message in the
        # text, we raise it. This handles any real RequestErrors, like
        # "Connection refused". We also raise any other exception so _request()
        # doesn't return it as a regular response.
        if isinstance(response_or_request_error, BaseException):
            raise response_or_request_error

    def _maybe_raise_error_from_json(self, response_or_request_error):
        # Get error message from ajax.php JSON Response or RequestError
        try:
            json = response_or_request_error.json()
        except errors.RequestError:
            # Response or RequestError is not JSON
            pass
        else:
            if (
                isinstance(json, dict)
                and json.get('Result') == 'Error'
                and json.get('Message')
            ):
                raise errors.RequestError(utils.html.as_text(json['Message']))

    def _maybe_raise_error_from_html(self, response_or_request_error):
        # Only attempt to find an error message if this looks like HTML. This
        # prevents a warning from bs4 about parsing non-HTML.
        text = str(response_or_request_error)
        if all(c in text for c in '<>\n'):
            doc = utils.html.parse(text)
            try:
                error_header_tag = doc.select('#content .page__title', string=re.compile(r'(?i:error)'))
                error_container_tag = error_header_tag[0].parent
                error_msg_tag = error_container_tag.find('div', attrs={'class': 'panel__body'})
                error_msg = error_msg_tag.get_text().strip()
                if error_msg:
                    raise errors.RequestError(error_msg)
            except (AttributeError, IndexError):
                # No error message found
                pass

    def _get_anti_csrf_token(self, response):
        json = response.json()
        try:
            return json['AntiCsrfToken']
        except KeyError as e:
            _log.debug('Failed to find "AntiCsrfToken": %r', response)
            raise errors.RequestError(
                'Looks like you have 2FA enabled? 2FA is not supported.\n'
                'Feel free to make a PR or report this as a bug if you see this with 2FA disabled.'
            ) from e

    async def _get_auth(self):
        assert self.is_logged_in
        response = await self._request('GET', self._base_url)
        doc = utils.html.parse(response)
        auth_regex = re.compile(r'logout\.php\?.*\bauth=([0-9a-fA-F]+)')
        logout_link_tag = doc.find('a', href=auth_regex)
        if logout_link_tag:
            logout_link_href = logout_link_tag['href']
            match = auth_regex.search(logout_link_href)
            return match.group(1)

        raise RuntimeError('Could not find auth')

    async def _login(self):
        if not self.options.get('username'):
            raise errors.RequestError('Login failed: No username configured')
        elif not self.options.get('password'):
            raise errors.RequestError('Login failed: No password configured')

        post_data = {
            'username': self.options['username'],
            'password': self.options['password'],
            'passkey': self._passkey,
            # 'keeplogged': '1',
        }
        response = await self._request(
            method='POST',
            url=self._ajax_url,
            params={'action': 'login'},
            data=post_data,
            error_prefix='Login failed',
        )
        self._anti_csrf_token = self._get_anti_csrf_token(response)

    async def _logout(self):
        try:
            await self._request(
                method='GET',
                url=self._logout_url,
                params={'auth': await self._get_auth()},
                error_prefix='Logout failed',
            )

        finally:
            if hasattr(self, '_anti_csrf_token'):
                delattr(self, '_anti_csrf_token')

    async def get_announce_url(self):
        return self._announce_url

    async def upload(self, tracker_jobs):
        post_data = tracker_jobs.post_data
        post_data['AntiCsrfToken'] = self._anti_csrf_token

        _log.debug('POSTing data:')
        for k, v in post_data.items():
            _log.debug(' * %s = %s', k, v)

        post_files = {
            'file_input': {
                'file': tracker_jobs.torrent_filepath,
                'mimetype': 'application/x-bittorrent',
            },
        }
        _log.debug('POSTing files: %r', post_files)

        response = await utils.http.post(
            url=self._upload_url,
            cache=False,
            user_agent=True,
            data=post_data,
            files=post_files,
            # Ignore the HTTP redirect (should be 302 Found) so we can get the
            # torrent URL from the "Location" response header
            follow_redirects=False,
        )

        # The server needs some time to process the uploaded metadata before the
        # torrent is registered by the tracker and we can seed it. If we don't
        # sleep() here, the torrent may be announced before the tracker knows it
        # exists.
        await asyncio.sleep(1)

        return self._handle_upload_response(response)

    def _handle_upload_response(self, response):
        # "Location" header should contain the uploaded torrent's URL
        _log.debug('Upload response headers: %r', response.headers)
        location = response.headers.get('Location')
        _log.debug('Upload response location: %r', location)
        if location:
            torrent_page_url = urllib.parse.urljoin(self.options['base_url'], location)
            # Redirect URL should start with "https://.../torrents.php"
            if torrent_page_url.startswith(self._torrents_url):
                return torrent_page_url

        # Find error message in HTML
        doc = utils.html.parse(response)

        # Find and raise error message
        alert_tag = doc.find(class_='alert')
        if alert_tag:
            msg = utils.html.as_text(alert_tag)
            raise errors.RequestError(f'Upload failed: {msg}')

        # Failed to find error message
        dump_filepath = 'ptp_upload_failed.html'
        utils.html.dump(response, dump_filepath)
        raise errors.RequestError(f'Failed to interpret response (see {dump_filepath})')

    def normalize_imdb_id(self, imdb_id):
        """
        Format IMDb ID for PTP

        PTP expects 7-characters, right-padded with "0" and without the leading
        "tt".

        If `imdb_id` is falsy (e.g. `None`, empty string, etc), or if it isn't
        an IMDb ID, return "0".
        """
        imdb_id = str(imdb_id)
        match = re.search(r'^(?:tt|)(\d+)$', imdb_id)
        if match:
            imdb_id_digits = match.group(1)
            if set(imdb_id_digits) != {'0'}:
                return match.group(1).rjust(7, '0')
        return '0'

    async def get_ptp_group_id_by_imdb_id(self, imdb_id):
        """
        Convert IMDb ID to PTP group ID

        Any :class:`~.RequestError` is caught and passed to
        :meth:`.TrackerBase.error`.

        :return: PTP group ID or `None` if PTP doesn't have a group for
            `imdb_id`
        """
        if imdb_id:
            _log.debug('%s: Fetching PTP group ID', imdb_id)
            try:
                await self.login()
                response = await self._request(
                    method='GET',
                    url=self._torrents_url,
                    params={
                        'imdb': self.normalize_imdb_id(imdb_id),
                        'json': '1',
                    },
                    cache=True,
                )

            except errors.RequestError as e:
                self.error(e)

            else:
                match = re.search(r'id=(\d+)', response.headers.get('location', ''))
                if match:
                    _log.debug('%s: PTP group ID: %s', imdb_id, match.group(1))
                    return match.group(1)
                else:
                    _log.debug('%s: No PTP group ID', imdb_id)

    async def get_movie_metadata(self, imdb_id):
        """
        Get metadata about movie as :class:`dict` from PTP website

        :param imdb_id: IMDb ID (e.g. ``tt123456``)

        The returned :class:`dict` always has the following keys:

            - ``countries`` (:class:`tuple` of :class:`str`)
            - ``languages`` (:class:`tuple` of :class:`str`)
            - ``plot`` (class:`str`)
            - ``poster`` (class:`str`, image URL)
            - ``tags`` (:class:`tuple` of :class:`str`)
            - ``title`` (class:`str`)
            - ``year`` (class:`str`)

        :raise RequestedNotFoundError: if no metadata is found.
        :raise RequestError: if the request fails.
        """
        _log.debug('%s: Fetching metadata from PTP', imdb_id)
        if imdb_id:
            await self.login()
            response = await self._request(
                method='GET',
                url=self._ajax_url,
                params={
                    'action': 'torrent_info',
                    'imdb': self.normalize_imdb_id(imdb_id),
                },
                cache=True,
            )

            # We expect a list of JSON objects that contains exactly one JSON
            # object.
            results = response.json()
            _log.debug('%s: Raw movie metadata: %r', imdb_id, response)
            assert len(results) == 1, results
            metadata = results[0]

            # If `imdb_id` is invalid/unknown, "title" is `null` in the JSON
            # response, which is translated to `None`.
            if not metadata.get('title'):
                raise errors.RequestedNotFoundError(imdb_id)

        else:
            metadata = {}

        def comma_split(string):
            lst = []
            for val in string.split(','):
                val = val.strip()
                if val:
                    lst.append(val)
            return lst

        return {
            'title': metadata.get('title') or '',
            'plot': metadata.get('plot') or '',
            'poster': metadata.get('art') or '',
            'year': metadata.get('year') or '',
            'tags': comma_split(metadata.get('tags', '')),
            'countries': comma_split(metadata.get('Countries', '')),
            'languages': comma_split(metadata.get('Languages', '')),
        }

    async def get_artist_metadata(self, artist):
        """
        Get metadata about artist as :class:`dict` from PTP website

        :param artist: Name, IMDb URL/ID (e.g. ``nm123456``) or PTP URL/ID

        :raise RequestedNotFoundError: if `artist` is unknown on both PTP and
            IMDb.
        :raise RequestError: if the metadata request fails.
        """
        _log.debug('%s: Fetching artist metadata from PTP', artist)
        await self.login()
        try:
            response = await self._request(
                method='POST',
                url=self._artist_url,
                data={
                    'action': 'find',
                    'name': (
                        # PTP accepts PTP artist URLs, but not a naked ID.
                        self._get_artist_url(artist)
                        if re.search(r'^\d+$', artist) else
                        artist
                    ),
                    'AntiCsrfToken': self._anti_csrf_token,
                },
                cache=True,
            )
        except errors.RequestError as e:
            if 'not found' in str(e):
                raise errors.RequestedNotFoundError(artist) from e
            else:
                raise e
        else:
            return self._get_artist_dict(response)

    async def create_artist(self, name):
        """
        Create artist on PTP

        .. warning:: This must only be called if the artist doesn't exist on
            IMDb or PTP.

        :param name: Complete canonical name of the artist

        :return: :class:`dict` with the keys: ``name``, ``id``, ``url``

        :raise RequestError: if the metadata request fails.
        """
        await self.login()
        _log.debug('Creating artist: %r', name)
        response = await self._request(
            method='POST',
            url=self._artist_url,
            data={
              'action': 'create',
              'name': name,
              'AntiCsrfToken': self._anti_csrf_token,
            },
            cache=False,
        )
        return self._get_artist_dict(response)

    def _get_artist_dict(self, response):
        _log.debug('Raw artist metadata: %r', response)
        # Raise RequestError if response is not valid JSON.
        artist = response.json()
        try:
            return {
                'name': artist['ArtistName'],
                'id': artist['ArtistId'],
                'url': self._get_artist_url(artist['ArtistId'])
            }
        except KeyError as e:
            if artist.get('Message'):
                raise errors.RequestError(artist['Message']) from e
            else:
                raise errors.RequestError(f'Unexpected response: {artist}') from e

    def _get_artist_url(self, artist_id):
        return self._artist_url + '?id=' + str(artist_id)

    @staticmethod
    def calculate_piece_size(bytes):
        """
        Return the recommended piece size for a given content size

        :param bytes: Torrent's content size without any excluded files
        """
        exponent = math.ceil(math.log2(bytes / 1050))
        # Allowed piece size range: 32 KiB ... 16 MiB
        exponent = max(15, min(24, exponent))
        return int(math.pow(2, exponent))

    @staticmethod
    def calculate_piece_size_min_max(bytes):
        """
        Return the allowed minimum and maximum piece size for a given
        content size

        :param bytes: torrent's content size without any excluded files

        :raise ValueError: if `bytes` is negative or otherwise unexpected
        """
        if bytes <= 0:
            raise ValueError(f'Unexpected size: {bytes!r}')

        # NOTE: The algorithm below is from the website's javascript.

        max_exponent = math.ceil(math.log2(bytes / 500))
        min_exponent = math.floor(math.log2(bytes / 2000))

        # 2^24 or 16MiB is the max for uTorrent 2.x
        min_exponent = min(24, min_exponent)
        max_exponent = min(24, max_exponent)

        # Compatibility for uTorrent 2.x creator which only supports up to 4 MiB piece sizes
        if min_exponent > 22 and bytes <= 128 * 1024 * 1024 * 1024:
            min_exponent = 22
        if min_exponent > 22 and bytes <= 256 * 1024 * 1024 * 1024:
            min_exponent = 23

        # Tiny torrents < ~40MiB need a sane lower bound
        min_exponent = max(15, min_exponent)
        max_exponent = max(18, max_exponent)

        return (
            math.pow(2, min_exponent),
            math.pow(2, max_exponent),
        )



"""
Dummy tracker for testing and debugging
"""

import asyncio
import functools
import os
import random
import string

from .. import errors, jobs, uis, utils
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


DummyImageHost = utils.types.ImageHost(allowed=('dummy',))


class DummyTrackerConfig(base.TrackerConfigBase):
    defaults = {
        'base_url': 'http://localhost',
        'username': '',
        'password': '',
        'exclude': (),
        'source': 'DMY',
        'image_host': utils.configfiles.config_value(
            value=utils.types.ListOf(
                item_type=DummyImageHost,
                default=DummyImageHost.options,
                separator=',',
            ),
            description=(
                'List of image hosting service names. The first service is normally used '
                + 'with the others as backup if uploading to the first fails.\n'
                + 'Supported services: ' + ', '.join(DummyImageHost.options)
            ),
        ),
    }

    argument_definitions = {
        'submit': {
            ('--imdb', '--im'): {
                'help': 'IMDb ID or URL',
                'type': utils.argtypes.webdb_id('imdb'),
            },
            ('--tmdb', '--tm'): {
                'help': 'TMDb ID or URL',
                'type': utils.argtypes.webdb_id('tmdb'),
            },
            ('--skip-category', '-C'): {
                'help': 'Do not ask for category',
                'action': 'store_true',
            },
            ('--screenshots-count', '--ssc'): {
                'help': 'How many screenshots to make',
                'type': utils.argtypes.number_of_screenshots(min=3, max=10),
            },
            ('--screenshots', '--ss'): {
                'help': (
                    'Path(s) to created screenshot file(s)\n'
                    'Directories are searched recursively.\n'
                    'Precreated screenshots are used in addition to automatically generated screenshots.'
                ),
                'nargs': '+',
                'action': 'extend',
                'type': utils.argtypes.files_with_extension('png'),
            },
            ('--delay', '-d'): {
                'help': 'Number of seconds login, upload and logout take each',
                'type': float,
                'default': 1.0,
            },
            ('--get-announce-from-website'): {
                'help': 'Whether to act like we are getting the announce URL from the website ',
                'action': 'store_true',
            },
            ('--poster'): {
                'help': 'Poster file or URL ',
            },
        },
        'torrent-create': {
            ('--delay', '-d'): {
                'help': 'Number of seconds login and logout take each',
                'type': float,
                'default': 0.0,
            },
            ('--get-announce-from-website'): {
                'help': 'Whether to act like we are getting the announce URL from the website ',
                'action': 'store_true',
            },
        },
    }


class DummyTrackerJobs(base.TrackerJobsBase):

    release_name_separator = '_'

    @functools.cached_property
    def jobs_before_upload(self):
        return (
            # Interactive jobs
            self.playlists_job,
            self.tmdb_job,
            self.imdb_job,
            self.release_name_job,
            self.category_job,
            self.scene_check_job,

            # Background jobs
            self.create_torrent_job,
            self.screenshots_job,
            self.upload_screenshots_job,
            self.poster_job,
            self.mediainfo_job,

            # Silly jobs
            self.random_string_job,
            self.stupid_question_job,
            self.say_something_job,
        )

    @functools.cached_property
    def random_string_job(self):
        return jobs.custom.CustomJob(
            name=self.get_job_name('random-string'),
            label='Random String',
            precondition=self.make_precondition('random_string_job'),
            worker=self.generate_random_string,
            catch=(
                errors.RequestError,
            ),
            **self.common_job_args(ignore_cache=True),
        )

    async def generate_random_string(self, job):
        # await asyncio.sleep(3)
        # self.random_string_job.warn('Watch out!')

        await asyncio.sleep(1)

        # raise errors.RequestError('foo :(')
        # raise RuntimeError('foo D:')

        return ''.join(
            random.choice(string.ascii_lowercase)
            for _ in range(30)
        )

    @functools.cached_property
    def stupid_question_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('stupid-question'),
            label='Stupid Question',
            precondition=self.make_precondition('stupid_question_job'),
            prejobs=(
                self.random_string_job,
            ),
            autodetect=self.autodetect_stupid_answer,
            question='Is the random string random enough?',
            options=(
                'Possibly',
                'Maybe',
                'Perhaps',
            ),
            multichoice=True,
            validate=self._validate_chosen,
            callbacks={
                'finished': lambda job: job.clear_warnings(),
            },
            **self.common_job_args(),
        )

    def _validate_chosen(self, chosen):
        labels = [option[0] for option in chosen]
        if 'Perhaps' in labels:
            raise ValueError('I have decided that "Perhaps" is not an option after all.')
        elif len(labels) < 2:
            raise ValueError('Pick at least 2 options.')

    async def autodetect_stupid_answer(self, job):
        assert self.random_string_job.is_finished

        # raise RuntimeError('How would I know?')

        random_string = self.random_string_job.output[0]
        if all(character in random_string for character in 'abc'):
            return 'Maybe'
        else:
            self.stupid_question_job.add_prompt(
                uis.prompts.RadioListPrompt(
                    question='Is this job too stupid?',
                    options=('Yes', 'No'),
                    callbacks=(
                        self.prompt_stupid_answer_callback,
                    ),
                )
            )
            self.stupid_question_job.add_prompt(
                uis.prompts.CheckListPrompt(
                    question='Are you sure?',
                    options=('Yes', 'No', 'Maybe?'),
                    callbacks=(
                        self.prompt_stupid_answer_callback_2,
                    ),
                )
            )
            self.stupid_question_job.add_prompt(
                uis.prompts.RadioListPrompt(
                    options=('Foo', 'Bar'),
                    callbacks=(),
                )
            )

    def prompt_stupid_answer_callback(self, result):
        if result == 'Yes':
            self.stupid_question_job.error('This job is too stupid.')
        else:
            self.stupid_question_job.warn('Choose wisely!')

    def prompt_stupid_answer_callback_2(self, result):
        if 'Yes' in result and 'No' in result:
            self.stupid_question_job.error('Yes and No? Be more decisive!')
        elif 'Maybe?' in result:
            self.stupid_question_job.warn('Okay?')
        elif 'Yes' not in result:
            self.stupid_question_job.error('This job may be too stupid after all!')

    @functools.cached_property
    def say_something_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('say-something'),
            label='Something',
            precondition=self.make_precondition('say_something_job'),
            prejobs=(
                self.random_string_job,
            ),
            nonfatal_exceptions=(
                ValueError,
            ),
            text=self.generate_something,
            **self.common_job_args(),
        )

    async def generate_something(self):
        assert self.random_string_job.is_finished
        word = self.random_string_job.output[0]

        await asyncio.sleep(6)
        self.say_something_job.warn('Still thinking...')
        await asyncio.sleep(6)

        self.say_something_job.clear_warnings()
        if 'x' in word:
            raise ValueError(f'There is an "x" in "{word}"! Yuck!')
        elif 'a' in word:
            return 'Avocado!'
        else:
            return f'This is a completely original random string: {word}'

    @functools.cached_property
    def category_job(self):
        if not self.options['skip_category']:
            return jobs.dialog.ChoiceJob(
                name=self.get_job_name('category'),
                label='Category',
                precondition=self.make_precondition('category_job'),
                options=(
                    (str(typ).capitalize(), typ)
                    for typ in utils.types.ReleaseType if typ
                ),
                autodetected=self.release_name.type,
                **self.common_job_args(),
            )


class DummyTracker(base.TrackerBase):
    name = 'dummy'
    label = 'DuMmY'

    @property
    def torrent_source_field(self):
        return self.options['source']

    setup_howto_template = (
        'This is just a no-op tracker for testing and demonstration.'
    )

    TrackerJobs = DummyTrackerJobs
    TrackerConfig = DummyTrackerConfig

    async def _login(self):
        _log.debug('%s: Logging in with %r', self.name, self.options)
        await asyncio.sleep(self.options['delay'])
        _log.debug('%s: Logged in with %r', self.name, self.options)

    async def _logout(self):
        _log.debug('%s: Logging out', self.name)
        await asyncio.sleep(self.options['delay'])
        _log.debug('%s: Logged out', self.name)

    async def get_announce_url(self):
        if self.options['get_announce_from_website']:
            _log.debug('%s: Getting announce URL from website', self.name)
            await self.login()
            await asyncio.sleep(1)
        else:
            _log.debug('%s: Getting announce URL from config file', self.name)
        return 'http://localhost:123/f1dd15718/announce'

    async def upload(self, tracker_jobs):
        if tracker_jobs.create_torrent_job.output:
            torrent_file = tracker_jobs.create_torrent_job.output[0]
        else:
            raise errors.RequestError('Torrent file was not created.')
        _log.debug('%s: Uploading %s', self.name, torrent_file)
        await asyncio.sleep(self.options['delay'])
        return f'http://localhost/{os.path.basename(torrent_file)}'



"""
Normalized representation of a sequence of :mod:`~.disc.bluray` or :mod:`~.disc.dvd` video files
"""

import functools

from ... import utils


class Playlist:
    """Representation of multiple videos in a sequence"""

    def __init__(self, items, filepath, *, discpath=None, duration=None, is_main=False):
        self._items = tuple(items)
        self._filepath = str(filepath)
        self._discpath = str(discpath) if discpath else None
        self._duration = duration
        self.is_main = is_main

    @functools.cached_property
    def type(self):
        """
        Underlying disc format

        "bluray" if :attr:`filename` has the file extension "mpls".
        "dvd" if :attr:`filename` has the file extension "ifo".
        Empty string otherwise.

        Matching is done case-insensitively.
        """
        file_extension = utils.fs.file_extension(self._filepath).lower()
        if file_extension == 'mpls':
            return 'bluray'
        elif file_extension == 'ifo':
            return 'dvd'
        else:
            return ''

    @functools.cached_property
    def label(self):
        """
        Human-readable string that serves as a unique playlist ID

        Combination of :attr:`discpath` (if provided) and :attr:`filename`.
        """
        if self.discname:
            return f'{self.discname}[{utils.fs.strip_extension(self.filename)}]'
        else:
            return utils.fs.strip_extension(self.filename)

    @property
    def items(self):
        """Sequence of file paths in the playlist"""
        return self._items

    @functools.cached_property
    def largest_item(self):
        """Largest file in :attr:`items`"""
        items_sorted = sorted(
            self.items,
            key=lambda filepath: (utils.fs.file_size(filepath) or 0),
            reverse=True,
        )
        return items_sorted[0]

    @property
    def filepath(self):
        """Path of the playlist file (e.g. MPLS or IFO file)"""
        return self._filepath

    @property
    def filename(self):
        """Basename of :attr:`filepath`"""
        return utils.fs.basename(self._filepath)

    @property
    def discpath(self):
        """
        Path of the directory that contains :attr:`filepath`

        Usually, this ia directory path that contains a "BDMV" or "VIDEO_TS" directory.
        """
        return self._discpath

    @property
    def discname(self):
        """Basename of :attr:`discpath`"""
        if self.discpath:
            return utils.fs.basename(self.discpath)

    @functools.cached_property
    def size(self):
        """Combined file size of all :attr:`items` as :attr:`utils.types.Bytes`"""
        return utils.types.Bytes(sum(
            # file_size() returns `None` by default.
            (utils.fs.file_size(item) or 0)
            for item in self.items
        ))

    @functools.cached_property
    def duration(self):
        """Combined runtime of :attr:`items` in seconds"""
        if self._duration is None:
            self._duration = sum(
                utils.mediainfo.get_duration(item)
                for item in self.items
            )
        return utils.types.Timestamp(self._duration)

    @property
    def is_main(self):
        """
        Simple boolean mark to track main playlists

        This property is mutable.
        """
        return self._is_main

    @is_main.setter
    def is_main(self, is_main):
        self._is_main = bool(is_main)

    # Attributes that make an instance unique.
    _id_attributes = ('discpath', 'filepath', 'items', 'duration')

    @property
    def id(self):
        """
        Sequence of values that make this instance unique

        Instances with the same :attr:`id` are equal.
        """
        return tuple(
            getattr(self, attr)
            for attr in self._id_attributes
        )

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.id == other.id
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return f'<{type(self).__name__} {self.label}>'


def mark_main_playlists(playlists):
    """Find the main playlist(s) and set their :attr:`~.Playlist.is_main` property to ``True``"""
    if not playlists:
        return

    # Find the largest playlist first. We go by size instead of runtime because the runtime is often
    # wrong while size always goes up as video complexity/quality increases, and the main playlist
    # is most likely of the highest complexity/quality.
    playlists_sorted = sorted(playlists, key=lambda playlist: playlist.size)
    largest_playlist = playlists_sorted[-1]

    # Main playlists have a similar duration as the largest playlist.
    duration_min = largest_playlist.duration * 0.7
    duration_max = largest_playlist.duration * 1.3
    for playlist in playlists:
        playlist.is_main = bool(duration_min <= playlist.duration <= duration_max)
